@@ -3,7 +3,7 @@
  * License, v. 2.0.  If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
- * Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.
+ * Copyright 1997 - July 2008 CWI, August 2008 - 2017 MonetDB B.V.
  */
 
 /*
@@ -208,11 +208,11 @@ GDKfdlocate(int farmid, const char *nme, const char *mode, const char *extension
 #ifdef WIN32
 	flags |= strchr(mode, 'b') ? O_BINARY : O_TEXT;
 #endif
-	fd = open(path, flags, MONETDB_MODE);
+	fd = open(path, flags | O_CLOEXEC, MONETDB_MODE);
 	if (fd < 0 && *mode == 'w') {
 		/* try to create the directory, in case that was the problem */
 		if (GDKcreatedir(path) == GDK_SUCCEED) {
-			fd = open(path, flags, MONETDB_MODE);
+			fd = open(path, flags | O_CLOEXEC, MONETDB_MODE);
 			if (fd < 0)
 				GDKsyserror("GDKfdlocate: cannot open file %s\n", path);
 		}
@@ -379,7 +379,7 @@ GDKextend(const char *fn, size_t size)
 	 * bytes without O_BINARY. */
 	flags |= O_BINARY;
 #endif
-	if ((fd = open(fn, flags)) >= 0) {
+	if ((fd = open(fn, flags | O_CLOEXEC)) >= 0) {
 		rt = GDKextendf(fd, size, fn);
 		close(fd);
 	} else {
@@ -577,7 +577,7 @@ GDKload(int farmid, const char *nme, const char *ext, size_t size, size_t *maxsi
  * physical @%home@ where they are to be located.  The subdirectories
  * BAT and PRG contain what its name says.
  *
- * A BAT created by @%BATnew@ is considered temporary until one calls
+ * A BAT created by @%COLnew@ is considered temporary until one calls
  * the routine @%BATsave@. This routine reserves disk space and checks
  * for name clashes.
  *
@@ -585,43 +585,35 @@ GDKload(int farmid, const char *nme, const char *ext, size_t size, size_t *maxsi
  * merely copies the data into place.  Failure to read or write the
  * BAT results in a NULL, otherwise it returns the BAT pointer.
  */
-static BATstore *
+static BAT *
 DESCload(int i)
 {
 	str s, nme = BBP_physical(i);
-	BATstore *bs;
 	BAT *b = NULL;
-	int ht, tt;
+	int tt;
 
 	IODEBUG {
 		fprintf(stderr, "#DESCload %s\n", nme ? nme : "<noname>");
 	}
-	bs = BBP_desc(i);
+	b = BBP_desc(i);
 
-	if (bs == NULL)
+	if (b == NULL)
 		return 0;
-	b = &bs->B;
 
-	ht = b->htype;
 	tt = b->ttype;
-	if ((ht < 0 && (ht = ATOMindex(s = ATOMunknown_name(ht))) < 0) ||
-	    (tt < 0 && (tt = ATOMindex(s = ATOMunknown_name(tt))) < 0)) {
+	if ((tt < 0 && (tt = ATOMindex(s = ATOMunknown_name(tt))) < 0)) {
 		GDKerror("DESCload: atom '%s' unknown, in BAT '%s'.\n", s, nme);
 		return NULL;
 	}
-	b->htype = ht;
 	b->ttype = tt;
-	b->H->hash = b->T->hash = NULL;
-	/* mil shouldn't mess with just loaded bats */
-	if (b->batStamp > 0)
-		b->batStamp = -b->batStamp;
+	b->thash = NULL;
 
 	/* reconstruct mode from BBP status (BATmode doesn't flush
 	 * descriptor, so loaded mode may be stale) */
 	b->batPersistence = (BBP_status(b->batCacheid) & BBPPERSISTENT) ? PERSISTENT : TRANSIENT;
 	b->batCopiedtodisk = 1;
 	DESCclean(b);
-	return bs;
+	return b;
 }
 
 void
@@ -630,12 +622,9 @@ DESCclean(BAT *b)
 	b->batDirtyflushed = DELTAdirty(b) ? TRUE : FALSE;
 	b->batDirty = 0;
 	b->batDirtydesc = 0;
-	b->H->heap.dirty = 0;
-	b->T->heap.dirty = 0;
-	if (b->H->vheap)
-		b->H->vheap->dirty = 0;
-	if (b->T->vheap)
-		b->T->vheap->dirty = 0;
+	b->theap.dirty = 0;
+	if (b->tvheap)
+		b->tvheap->dirty = 0;
 }
 
 /* spawning the background msync should be done carefully 
@@ -648,6 +637,7 @@ DESCclean(BAT *b)
 #define MSYNC_BACKGROUND
 
 #ifndef DISABLE_MSYNC
+#ifndef MS_ASYNC
 struct msync {
 	bat id;
 	Heap *h;
@@ -657,63 +647,71 @@ static void
 BATmsyncImplementation(void *arg)
 {
 	Heap *h = ((struct msync *) arg)->h;
-	char *adr;
-	size_t len;
-	size_t offset;
 
-	adr = h->base;
-	offset = ((size_t) adr % MT_pagesize());
-	len = MT_pagesize() * (1 + ((h->base + h->free - adr) / MT_pagesize()));
-	if (offset)
-		adr -= MT_pagesize() - offset;
-	if (len)
-		(void) MT_msync(adr, len);
+	(void) MT_msync(h->base, h->size);
 	BBPunfix(((struct msync *) arg)->id);
 	GDKfree(arg);
 }
 #endif
+#endif
 
 void
 BATmsync(BAT *b)
 {
+	/* we don't sync views */
+	if (isVIEW(b))
+		return;
+	/* we don't sync transients */
+	if (b->theap.farmid != 0 ||
+	    (b->tvheap != NULL && b->tvheap->farmid != 0))
+		return;
 #ifndef DISABLE_MSYNC
-#ifdef MSYNC_BACKGROUND
-	MT_Id tid;
-#endif
-	struct msync *arg;
-
-	assert(b->batPersistence == PERSISTENT);
-	if (b->T->heap.storage == STORE_MMAP &&
-	    (arg = GDKmalloc(sizeof(*arg))) != NULL) {
-		arg->id = b->batCacheid;
-		arg->h = &b->T->heap;
-		BBPfix(b->batCacheid);
-#ifdef MSYNC_BACKGROUND
-		if (MT_create_thread(&tid, BATmsyncImplementation, arg, MT_THR_DETACHED) < 0) {
-			/* don't bother if we can't create a thread */
-			BBPunfix(b->batCacheid);
-			GDKfree(arg);
-		}
+#ifdef MS_ASYNC
+	if (b->theap.storage == STORE_MMAP)
+		(void) msync(b->theap.base, b->theap.free, MS_ASYNC);
+	if (b->tvheap && b->tvheap->storage == STORE_MMAP)
+		(void) msync(b->tvheap->base, b->tvheap->free, MS_ASYNC);
 #else
-		BATmsyncImplementation(arg);
-#endif
-	}
-
-	if (b->T->vheap && b->T->vheap->storage == STORE_MMAP &&
-	    (arg = GDKmalloc(sizeof(*arg))) != NULL) {
-		arg->id = b->batCacheid;
-		arg->h = b->T->vheap;
-		BBPfix(b->batCacheid);
+	{
 #ifdef MSYNC_BACKGROUND
-		if (MT_create_thread(&tid, BATmsyncImplementation, arg, MT_THR_DETACHED) < 0) {
-			/* don't bother if we can't create a thread */
-			BBPunfix(b->batCacheid);
-			GDKfree(arg);
-		}
-#else
-		BATmsyncImplementation(arg);
+		MT_Id tid;
 #endif
+		struct msync *arg;
+
+		assert(b->batPersistence == PERSISTENT);
+		if (b->theap.storage == STORE_MMAP &&
+		    (arg = GDKmalloc(sizeof(*arg))) != NULL) {
+			arg->id = b->batCacheid;
+			arg->h = &b->theap;
+			BBPfix(b->batCacheid);
+#ifdef MSYNC_BACKGROUND
+			if (MT_create_thread(&tid, BATmsyncImplementation, arg, MT_THR_DETACHED) < 0) {
+				/* don't bother if we can't create a thread */
+				BBPunfix(b->batCacheid);
+				GDKfree(arg);
+			}
+#else
+			BATmsyncImplementation(arg);
+#endif
+		}
+
+		if (b->tvheap && b->tvheap->storage == STORE_MMAP &&
+		    (arg = GDKmalloc(sizeof(*arg))) != NULL) {
+			arg->id = b->batCacheid;
+			arg->h = b->tvheap;
+			BBPfix(b->batCacheid);
+#ifdef MSYNC_BACKGROUND
+			if (MT_create_thread(&tid, BATmsyncImplementation, arg, MT_THR_DETACHED) < 0) {
+				/* don't bother if we can't create a thread */
+				BBPunfix(b->batCacheid);
+				GDKfree(arg);
+			}
+#else
+			BATmsyncImplementation(arg);
+#endif
+		}
 	}
+#endif
 #else
 	(void) b;
 #endif	/* DISABLE_MSYNC */
@@ -724,85 +722,51 @@ BATsave(BAT *bd)
 {
 	gdk_return err = GDK_SUCCEED;
 	char *nme;
-	BATstore bs;
+	BAT bs;
 	BAT *b = bd;
 
 	BATcheck(b, "BATsave", GDK_FAIL);
 
+	assert(b->batCacheid > 0);
 	/* views cannot be saved, but make an exception for
 	 * force-remapped views */
 	if (isVIEW(b) &&
-	    !(b->H->heap.copied && b->H->heap.storage == STORE_MMAP) &&
-	    !(b->T->heap.copied && b->T->heap.storage == STORE_MMAP)) {
-		GDKerror("BATsave: %s is a view on %s; cannot be saved\n", BATgetId(b), VIEWhparent(b) ? BBPname(VIEWhparent(b)) : BBPname(VIEWtparent(b)));
+	    !(b->theap.copied && b->theap.storage == STORE_MMAP)) {
+		GDKerror("BATsave: %s is a view on %s; cannot be saved\n", BATgetId(b), BBPname(VIEWtparent(b)));
 		return GDK_FAIL;
 	}
 	if (!BATdirty(b)) {
 		return GDK_SUCCEED;
 	}
-	if (b->batCacheid < 0) {
-		b = BATmirror(b);
-	}
-	if (!DELTAdirty(b))
-		ALIGNcommit(b);
-	if (!b->halign)
-		b->halign = OIDnew(1);
-	if (!b->talign)
-		b->talign = OIDnew(1);
 
 	/* copy the descriptor to a local variable in order to let our
 	 * messing in the BAT descriptor not affect other threads that
 	 * only read it. */
 	bs = *BBP_desc(b->batCacheid);
 	/* fix up internal pointers */
-	b = &bs.BM;		/* first the mirror */
-	b->S = &bs.S;
-	b->H = &bs.T;
-	b->T = &bs.H;
-	b = &bs.B;		/* then the unmirrored version */
-	b->S = &bs.S;
-	b->H = &bs.H;
-	b->T = &bs.T;
+	b = &bs;
 
-	if (b->H->vheap) {
-		b->H->vheap = (Heap *) GDKmalloc(sizeof(Heap));
-		if (b->H->vheap == NULL)
-			return GDK_FAIL;
-		*b->H->vheap = *bd->H->vheap;
-	}
-	if (b->T->vheap) {
-		b->T->vheap = (Heap *) GDKmalloc(sizeof(Heap));
-		if (b->T->vheap == NULL) {
-			if (b->H->vheap)
-				GDKfree(b->H->vheap);
+	if (b->tvheap) {
+		b->tvheap = (Heap *) GDKmalloc(sizeof(Heap));
+		if (b->tvheap == NULL) {
 			return GDK_FAIL;
 		}
-		*b->T->vheap = *bd->T->vheap;
+		*b->tvheap = *bd->tvheap;
 	}
 
 	/* start saving data */
 	nme = BBP_physical(b->batCacheid);
-	if (b->batCopiedtodisk == 0 || b->batDirty || b->H->heap.dirty)
-		if (err == GDK_SUCCEED && b->htype)
-			err = HEAPsave(&b->H->heap, nme, "head");
-	if (b->batCopiedtodisk == 0 || b->batDirty || b->T->heap.dirty)
+	if (b->batCopiedtodisk == 0 || b->batDirty || b->theap.dirty)
 		if (err == GDK_SUCCEED && b->ttype)
-			err = HEAPsave(&b->T->heap, nme, "tail");
-	if (b->H->vheap && (b->batCopiedtodisk == 0 || b->batDirty || b->H->vheap->dirty))
-		if (b->htype && b->hvarsized) {
-			if (err == GDK_SUCCEED)
-				err = HEAPsave(b->H->vheap, nme, "hheap");
-		}
-	if (b->T->vheap && (b->batCopiedtodisk == 0 || b->batDirty || b->T->vheap->dirty))
+			err = HEAPsave(&b->theap, nme, "tail");
+	if (b->tvheap && (b->batCopiedtodisk == 0 || b->batDirty || b->tvheap->dirty))
 		if (b->ttype && b->tvarsized) {
 			if (err == GDK_SUCCEED)
-				err = HEAPsave(b->T->vheap, nme, "theap");
+				err = HEAPsave(b->tvheap, nme, "theap");
 		}
 
-	if (b->H->vheap)
-		GDKfree(b->H->vheap);
-	if (b->T->vheap)
-		GDKfree(b->T->vheap);
+	if (b->tvheap)
+		GDKfree(b->tvheap);
 
 	if (err == GDK_SUCCEED) {
 		bd->batCopiedtodisk = 1;
@@ -817,102 +781,56 @@ BATsave(BAT *bd)
  * TODO: move to gdk_bbp.mx
  */
 BAT *
-BATload_intern(bat i, int lock)
+BATload_intern(bat bid, int lock)
 {
-	bat bid = abs(i);
-	str nme = BBP_physical(bid);
-	BATstore *bs = DESCload(bid);
+	str nme;
 	BAT *b;
 
-	if (bs == NULL) {
+	assert(bid > 0);
+
+	nme = BBP_physical(bid);
+	b = DESCload(bid);
+
+	if (b == NULL) {
 		return NULL;
 	}
-	b = &bs->B;
 
 	/* LOAD bun heap */
-	if (b->htype != TYPE_void) {
-		if (HEAPload(&b->H->heap, nme, "head", b->batRestricted == BAT_READ) != GDK_SUCCEED) {
-			return NULL;
-		}
-		assert(b->H->heap.size >> b->H->shift <= BUN_MAX);
-		b->batCapacity = (BUN) (b->H->heap.size >> b->H->shift);
-	} else {
-		b->H->heap.base = NULL;
-	}
 	if (b->ttype != TYPE_void) {
-		if (HEAPload(&b->T->heap, nme, "tail", b->batRestricted == BAT_READ) != GDK_SUCCEED) {
-			HEAPfree(&b->H->heap, 0);
+		if (HEAPload(&b->theap, nme, "tail", b->batRestricted == BAT_READ) != GDK_SUCCEED) {
+			HEAPfree(&b->theap, 0);
 			return NULL;
 		}
-		if (b->htype == TYPE_void) {
-			assert(b->T->heap.size >> b->T->shift <= BUN_MAX);
-			b->batCapacity = (BUN) (b->T->heap.size >> b->T->shift);
-		}
-		if (b->batCapacity != (b->T->heap.size >> b->T->shift)) {
-			BUN cap = b->batCapacity;
-			gdk_return h;
-			if (cap < (b->T->heap.size >> b->T->shift)) {
-				cap = (BUN) (b->T->heap.size >> b->T->shift);
-				HEAPDEBUG fprintf(stderr, "#HEAPextend in BATload_inter %s " SZFMT " " SZFMT "\n", b->H->heap.filename, b->H->heap.size, headsize(b, cap));
-				h = HEAPextend(&b->H->heap, headsize(b, cap), b->batRestricted == BAT_READ);
-				b->batCapacity = cap;
-			} else {
-				HEAPDEBUG fprintf(stderr, "#HEAPextend in BATload_intern %s " SZFMT " " SZFMT "\n", b->T->heap.filename, b->T->heap.size, tailsize(b, cap));
-				h = HEAPextend(&b->T->heap, tailsize(b, cap), b->batRestricted == BAT_READ);
-			}
-			if (h != GDK_SUCCEED) {
-				HEAPfree(&b->H->heap, 0);
-				HEAPfree(&b->T->heap, 0);
-				return NULL;
-			}
-		}
+		assert(b->theap.size >> b->tshift <= BUN_MAX);
+		b->batCapacity = (BUN) (b->theap.size >> b->tshift);
 	} else {
-		b->T->heap.base = NULL;
-	}
-
-	/* LOAD head heap */
-	if (ATOMvarsized(b->htype)) {
-		if (HEAPload(b->H->vheap, nme, "hheap", b->batRestricted == BAT_READ) != GDK_SUCCEED) {
-			HEAPfree(&b->H->heap, 0);
-			HEAPfree(&b->T->heap, 0);
-			return NULL;
-		}
-		if (ATOMstorage(b->htype) == TYPE_str) {
-			strCleanHash(b->H->vheap, FALSE);	/* ensure consistency */
-		}
+		b->theap.base = NULL;
 	}
 
 	/* LOAD tail heap */
 	if (ATOMvarsized(b->ttype)) {
-		if (HEAPload(b->T->vheap, nme, "theap", b->batRestricted == BAT_READ) != GDK_SUCCEED) {
-			if (b->H->vheap)
-				HEAPfree(b->H->vheap, 0);
-			HEAPfree(&b->H->heap, 0);
-			HEAPfree(&b->T->heap, 0);
+		if (HEAPload(b->tvheap, nme, "theap", b->batRestricted == BAT_READ) != GDK_SUCCEED) {
+			HEAPfree(&b->theap, 0);
+			HEAPfree(b->tvheap, 0);
 			return NULL;
 		}
 		if (ATOMstorage(b->ttype) == TYPE_str) {
-			strCleanHash(b->T->vheap, FALSE);	/* ensure consistency */
+			strCleanHash(b->tvheap, FALSE);	/* ensure consistency */
 		}
 	}
 
 	/* initialize descriptor */
 	b->batDirtydesc = FALSE;
-	b->H->heap.parentid = b->T->heap.parentid = 0;
+	b->theap.parentid = 0;
 
 	/* load succeeded; register it in BBP */
-	BBPcacheit(bs, lock);
-
-	if (!DELTAdirty(b)) {
-		ALIGNcommit(b);
+	if (BBPcacheit(b, lock) != GDK_SUCCEED) {
+		HEAPfree(&b->theap, 0);
+		if (b->tvheap)
+			HEAPfree(b->tvheap, 0);
+		return NULL;
 	}
-
-	if ((b->batRestricted == BAT_WRITE && (GDKdebug & CHECKMASK)) ||
-	    (GDKdebug & PROPMASK)) {
-		++b->batSharecnt;
-		--b->batSharecnt;
-	}
-	return (i < 0) ? BATmirror(b) : b;
+	return b;
 }
 
 /*
@@ -933,48 +851,32 @@ BATload_intern(bat i, int lock)
 void
 BATdelete(BAT *b)
 {
-	bat bid = abs(b->batCacheid);
+	bat bid = b->batCacheid;
 	str o = BBP_physical(bid);
 	BAT *loaded = BBP_cache(bid);
 
+	assert(bid > 0);
 	if (loaded) {
 		b = loaded;
 		HASHdestroy(b);
 		IMPSdestroy(b);
+		OIDXdestroy(b);
 	}
-	assert(!b->H->heap.base || !b->T->heap.base || b->H->heap.base != b->T->heap.base);
-	if (b->batCopiedtodisk || (b->H->heap.storage != STORE_MEM)) {
-		if (b->htype != TYPE_void &&
-		    HEAPdelete(&b->H->heap, o, "head") &&
-		    b->batCopiedtodisk)
-			IODEBUG fprintf(stderr, "#BATdelete(%s): bun heap\n", BATgetId(b));
-	} else if (b->H->heap.base) {
-		HEAPfree(&b->H->heap, 1);
-	}
-	if (b->batCopiedtodisk || (b->T->heap.storage != STORE_MEM)) {
+	if (b->batCopiedtodisk || (b->theap.storage != STORE_MEM)) {
 		if (b->ttype != TYPE_void &&
-		    HEAPdelete(&b->T->heap, o, "tail") &&
+		    HEAPdelete(&b->theap, o, "tail") &&
 		    b->batCopiedtodisk)
 			IODEBUG fprintf(stderr, "#BATdelete(%s): bun heap\n", BATgetId(b));
-	} else if (b->T->heap.base) {
-		HEAPfree(&b->T->heap, 1);
+	} else if (b->theap.base) {
+		HEAPfree(&b->theap, 1);
 	}
-	if (b->H->vheap) {
-		assert(b->H->vheap->parentid == bid);
-		if (b->batCopiedtodisk || (b->H->vheap->storage != STORE_MEM)) {
-			if (HEAPdelete(b->H->vheap, o, "hheap") && b->batCopiedtodisk)
-				IODEBUG fprintf(stderr, "#BATdelete(%s): head heap\n", BATgetId(b));
-		} else {
-			HEAPfree(b->H->vheap, 1);
-		}
-	}
-	if (b->T->vheap) {
-		assert(b->T->vheap->parentid == bid);
-		if (b->batCopiedtodisk || (b->T->vheap->storage != STORE_MEM)) {
-			if (HEAPdelete(b->T->vheap, o, "theap") && b->batCopiedtodisk)
+	if (b->tvheap) {
+		assert(b->tvheap->parentid == bid);
+		if (b->batCopiedtodisk || (b->tvheap->storage != STORE_MEM)) {
+			if (HEAPdelete(b->tvheap, o, "theap") && b->batCopiedtodisk)
 				IODEBUG fprintf(stderr, "#BATdelete(%s): tail heap\n", BATgetId(b));
 		} else {
-			HEAPfree(b->T->vheap, 1);
+			HEAPfree(b->tvheap, 1);
 		}
 	}
 	b->batCopiedtodisk = FALSE;
@@ -1042,7 +944,7 @@ BATprintcolumns(stream *s, int argc, BAT *argv[])
 	for (n = 0, cnt = BATcount(argv[0]); n < cnt; n++) {
 		mnstr_write(s, "[ ", 1, 2);
 		for (i = 0; i < argc; i++) {
-			len = colinfo[i].s(&buf, &buflen, BUNtail(colinfo[i].i, BUNfirst(argv[i]) + n));
+			len = colinfo[i].s(&buf, &buflen, BUNtail(colinfo[i].i, n));
 			if (i > 0)
 				mnstr_write(s, ",\t", 1, 2);
 			mnstr_write(s, buf, 1, len);
@@ -1057,7 +959,7 @@ BATprintcolumns(stream *s, int argc, BAT *argv[])
 }
 
 gdk_return
-BATprintf(stream *s, BAT *b)
+BATprint(BAT *b)
 {
 	BAT *argv[2];
 	gdk_return ret = GDK_FAIL;
@@ -1065,16 +967,10 @@ BATprintf(stream *s, BAT *b)
 	argv[0] = BATdense(b->hseqbase, b->hseqbase, BATcount(b));
 	argv[1] = b;
 	if (argv[0] && argv[1]) {
-		BATroles(argv[0], NULL, b->hident);
-		ret = BATprintcolumns(s, 2, argv);
+		BATroles(argv[0], "h");
+		ret = BATprintcolumns(GDKstdout, 2, argv);
 	}
 	if (argv[0])
 		BBPunfix(argv[0]->batCacheid);
 	return ret;
 }
-
-gdk_return
-BATprint(BAT *b)
-{
-	return BATprintf(GDKstdout, b);
-}
