@@ -3,7 +3,7 @@
  * License, v. 2.0.  If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
- * Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.
+ * Copyright 1997 - July 2008 CWI, August 2008 - 2017 MonetDB B.V.
  */
 
 /*
@@ -21,18 +21,26 @@
 Symbol newFunction(str mod, str nme,int kind){
 	Symbol s;
 	InstrPtr p;
+	int varid;
 
 	s = newSymbol(nme,kind);
 	if (s == NULL)
 		return NULL;
-	p = newInstruction(NULL,kind);
+
+	varid = newVariable(s->def,nme,strlen(nme),TYPE_any);
+	if( varid < 0){
+		freeSymbol(s);
+		return NULL;
+	}
+
+	p = newInstruction(NULL, mod, nme);
 	if (p == NULL) {
 		freeSymbol(s);
 		return NULL;
 	}
-	setModuleId(p, mod);
-	setFunctionId(p, nme);
-	setDestVar(p, newVariable(s->def,GDKstrdup(nme),TYPE_any));
+	p->token = kind;
+	p->barrier = 0;
+	setDestVar(p, varid);
 	pushInstruction(s->def,p);
 	return s;
 }
@@ -46,9 +54,9 @@ Symbol  getFunctionSymbol(Module scope, InstrPtr p){
 	Module m;
 	Symbol s;
 
-	for(m= findModule(scope,getModuleId(p)); m; m= m->outer)
+	for(m= findModule(scope,getModuleId(p)); m; m= m->link)
 		if(idcmp(m->name, getModuleId(p))==0 ) {
-			s= m->subscope[(int)(getSubScope(getFunctionId(p)))];
+			s= m->space[getSymbolIndex(getFunctionId(p))];
 			for(; s; s= s->peer)
 				if( getSignature(s)->fcn == p->fcn)
 					return s;
@@ -77,7 +85,7 @@ void chkFlow(stream *out, MalBlkPtr mb)
 	int  var[DEPTH];
 	InstrPtr stmt[DEPTH];
 	int btop=0;
-	int retseen=0, yieldseen=0;
+	int endseen=0, retseen=0, yieldseen=0;
 	int fixed=1;
 	InstrPtr p;
 
@@ -171,6 +179,7 @@ void chkFlow(stream *out, MalBlkPtr mb)
 			}
 			yieldseen= TRUE;
 			 }
+			/* fall through */
 		case RETURNsymbol:
 			{
 				InstrPtr ps = getInstrPtr(mb, 0);
@@ -200,6 +209,9 @@ void chkFlow(stream *out, MalBlkPtr mb)
 			break;
 	    case RAISEsymbol:
 	        break;
+	    case ENDsymbol:
+			endseen =1;
+	        break;
 		default:
 			if( isaSignature(p) ){
 				if( p->token == REMsymbol){
@@ -217,10 +229,11 @@ void chkFlow(stream *out, MalBlkPtr mb)
 		showScriptException(out, mb,lastInstruction,SYNTAX,
 			"instructions after END");
 #ifdef DEBUG_MAL_FCN
-		printFunction(out, mb, 0, LIST_MAL_ALL);
+		fprintFunction(stderr, mb, 0, LIST_MAL_ALL);
 #endif
 		mb->errors++;
 	}
+	if( endseen)
 	for(btop--; btop>=0;btop--){
 		showScriptException(out, mb,lastInstruction, SYNTAX,
 			"barrier '%s' without exit in %s[%d]",
@@ -276,13 +289,13 @@ static void replaceTypeVar(MalBlkPtr mb, InstrPtr p, int v, malType t){
 	int j,i,x,y;
 #ifdef DEBUG_MAL_FCN
 	char *tpenme = getTypeName(t);
-	mnstr_printf(GDKout,"replace type _%d by type %s\n",v, tpenme);
+	fprintf(stderr,"#replace type _%d by type %s\n",v, tpenme);
 	GDKfree(tpenme);
 #endif
 	for(j=0; j<mb->stop; j++){
 	    p= getInstrPtr(mb,j);
 #ifdef DEBUG_MAL_FCN
-		printInstruction(GDKout,mb,0,p,LIST_MAL_ALL);
+		fprintInstruction(stderr,mb,0,p,LIST_MAL_ALL);
 #endif
 	if( p->polymorphic)
 	for(i=0;i<p->argc; i++)
@@ -290,28 +303,28 @@ static void replaceTypeVar(MalBlkPtr mb, InstrPtr p, int v, malType t){
 		if( isaBatType(x)){
 			int tail;
 			int tx;
-			tail = getColumnType(x);
-			tx = getColumnIndex(x);
+			tail = getBatType(x);
+			tx = getTypeIndex(x);
 			if(v && tx == v && tail == TYPE_any){
 			    tx= 0;
 			    tail = t;
 			}
-			y= newBatType(TYPE_void,tail);
-			setAnyColumnIndex(y,tx);
+			y= newBatType(tail);
+			setTypeIndex(y,tx);
 			setArgType(mb,p,i,y);
 #ifdef DEBUG_MAL_FCN
 			{
 				char *xnme = getTypeName(x), *ynme = getTypeName(y);
-				mnstr_printf(GDKout," %d replaced %s->%s \n",i,xnme,ynme);
+				fprintf(stderr," %d replaced %s->%s \n",i,xnme,ynme);
 				GDKfree(xnme);
 				GDKfree(ynme);
 			}
 #endif
 		} else
-		if(getColumnIndex(x) == v){
+		if(getTypeIndex(x) == v){
 #ifdef DEBUG_MAL_FCN
 			char *xnme = getTypeName(x);
-			mnstr_printf(GDKout," replace x= %s polymorphic\n",xnme);
+			fprintf(stderr," replace x= %s polymorphic\n",xnme);
 			GDKfree(xnme);
 #endif
 			setArgType(mb,p,i,t);
@@ -319,13 +332,13 @@ static void replaceTypeVar(MalBlkPtr mb, InstrPtr p, int v, malType t){
 #ifdef DEBUG_MAL_FCN
 		else {
 			char *xnme = getTypeName(x);
-			mnstr_printf(GDKout," non x= %s %d\n",xnme,getColumnIndex(x));
+			fprintf(stderr," non x= %s %d\n",xnme,getTypeIndex(x));
 			GDKfree(xnme);
 		}
 #endif
 	}
 #ifdef DEBUG_MAL_FCN
-		printInstruction(GDKout,mb,0,p,LIST_MAL_ALL);
+		fprintInstruction(stderr,mb,0,p,LIST_MAL_ALL);
 #endif
 	}
 }
@@ -346,14 +359,14 @@ insertSymbolBefore(Module scope, Symbol prg, Symbol before)
 		if (c)
 			scope = c;
 	}
-	t = getSubScope(getFunctionId(sig));
-	assert(scope->subscope != NULL);
-	assert(scope->subscope[t] != NULL);
-	s = scope->subscope[t];
+	t = getSymbolIndex(getFunctionId(sig));
+	assert(scope->space != NULL);
+	assert(scope->space[t] != NULL);
+	s = scope->space[t];
 	prg->skip = before->skip;
 	prg->peer = before;
 	if (s == before) {
-		scope->subscope[t] = prg;
+		scope->space[t] = prg;
 	} else {
 		for (;;) {
 			assert(s != NULL);
@@ -381,17 +394,21 @@ cloneFunction(stream *out, Module scope, Symbol proc, MalBlkPtr mb, InstrPtr p)
 	InstrPtr pp;
 
 #ifdef DEBUG_CLONE
-	mnstr_printf(out,"clone the function %s to scope %s\n",
+	fprintf(stderr,"clone the function %s to scope %s\n",
 				 proc->name,scope->name);
-	printInstruction(out,mb,0,p,LIST_MAL_ALL);
+	fprintInstruction(stderr,mb,0,p,LIST_MAL_ALL);
 #endif
 	new = newFunction(scope->name, proc->name, getSignature(proc)->token);
+	if( new == NULL){
+		fprintf(stderr,"cloneFunction() failed");
+		return NULL;
+	}
 	freeMalBlk(new->def);
 	new->def = copyMalBlk(proc->def);
 	/* now change the definition of the original proc */
 #ifdef DEBUG_CLONE
-	mnstr_printf(out, "CLONED VERSION\n");
-	printFunction(out, new->def, 0, LIST_MAL_ALL);
+	fprintf(stderr, "CLONED VERSION\n");
+	fprintFunction(stderr, new->def, 0, LIST_MAL_ALL);
 #endif
 	/* check for errors after fixation , TODO*/
 	pp = getSignature(new);
@@ -402,15 +419,15 @@ cloneFunction(stream *out, Module scope, Symbol proc, MalBlkPtr mb, InstrPtr p)
 			if (v == TYPE_any)
 				replaceTypeVar(new->def, pp, v, t);
 			if (isaBatType(v)) {
-				if (getColumnIndex(v))
-					replaceTypeVar(new->def, pp, getColumnIndex(v), getColumnType(t));
+				if (getTypeIndex(v))
+					replaceTypeVar(new->def, pp, getTypeIndex(v), getBatType(t));
 			} else
-				replaceTypeVar(new->def, pp, getColumnIndex(v), t);
+				replaceTypeVar(new->def, pp, getTypeIndex(v), t);
 		}
 #ifdef DEBUG_MAL_FCN
 		else {
 			char *tpenme = getTypeName(v);
-			mnstr_printf(out,"%d remains %s\n", i, tpenme);
+			fprintf(stderr,"%d remains %s\n", i, tpenme);
 			GDKfree(tpenme);
 		}
 #endif
@@ -427,8 +444,8 @@ cloneFunction(stream *out, Module scope, Symbol proc, MalBlkPtr mb, InstrPtr p)
 		clrVarFixed(new->def, i);
 
 #ifdef DEBUG_MAL_FCN
-	mnstr_printf(out, "FUNCTION TO BE CHECKED\n");
-	printFunction(out, new->def, 0, LIST_MAL_ALL);
+	fprintf(stderr, "FUNCTION TO BE CHECKED\n");
+	fprintFunction(stderr, new->def, 0, LIST_MAL_ALL);
 #endif
 
 	/* check for errors after fixation , TODO*/
@@ -439,12 +456,12 @@ cloneFunction(stream *out, Module scope, Symbol proc, MalBlkPtr mb, InstrPtr p)
 			showScriptException(out, new->def, 0, MAL,
 								"Error in cloned function");
 #ifdef DEBUG_MAL_FCN
-			printFunction(out, new->def, 0, LIST_MAL_ALL);
+			fprintFunction(stderr, new->def, 0, LIST_MAL_ALL);
 #endif
 		}
 	}
 #ifdef DEBUG_CLONE
-	mnstr_printf(out, "newly cloned function added to %s %d \n",
+	fprintf(stderr, "newly cloned function added to %s %d \n",
 				 scope->name, i);
 	printFunction(out, new->def, 0, LIST_MAL_ALL);
 #endif
@@ -476,7 +493,7 @@ debugFunction(stream *fd, MalBlkPtr mb, MalStkPtr stk, int flg, int first, int s
 			if (p->token == REMsymbol)
 				mnstr_printf(fd,"%-40s\n",ps);
 			else {
-				mnstr_printf(fd,"%-40s\t#[%d] ("BUNFMT") %s ",ps, i, getRowCnt(mb,getArg(p,0)), (p->blk && p->blk->binding? p->blk->binding:""));
+				mnstr_printf(fd,"%-40s\t#[%d] ("BUNFMT") %s ",ps, i, getRowCnt(mb,getArg(p,0)), (p->blk? p->blk->binding:""));
 				for(j =0; j < p->retc; j++)
 					mnstr_printf(fd,"%d ",getArg(p,j));
 				if( p->argc - p->retc > 0)
@@ -486,7 +503,7 @@ debugFunction(stream *fd, MalBlkPtr mb, MalStkPtr stk, int flg, int first, int s
 				mnstr_printf(fd,"\n");
 			}
 			GDKfree(ps);
-		}
+		} else mnstr_printf(fd,"#failed instruction2str()\n");
 	}
 }
 
@@ -517,7 +534,7 @@ listFunction(stream *fd, MalBlkPtr mb, MalStkPtr stk, int flg, int first, int si
 				if (l > len)
 					len = l;
 				GDKfree(ps);
-			}
+			} else mnstr_printf(fd,"#failed instruction2str()\n");
 		}
 		mnstr_printf(fd, "%% " SZFMT " # length\n", len);
 	}
@@ -527,26 +544,60 @@ listFunction(stream *fd, MalBlkPtr mb, MalStkPtr stk, int flg, int first, int si
 
 void printFunction(stream *fd, MalBlkPtr mb, MalStkPtr stk, int flg)
 {
+	int i,j;
+	InstrPtr p;
+	// Set the used bits properly
+	for(i=0; i< mb->vtop; i++)
+		clrVarUsed(mb,i);
+	for(i=0; i< mb->stop; i++){
+		p= getInstrPtr(mb,i);
+		for(j= p->retc; j<p->argc; j++)
+			setVarUsed(mb, getArg(p,j));
+		if( p->barrier)
+			for(j= 0; j< p->retc; j++)
+				setVarUsed(mb, getArg(p,j));
+	}
 	listFunction(fd,mb,stk,flg,0,mb->stop);
 }
 
-Lifespan
-setLifespan(MalBlkPtr mb)
+void fprintFunction(FILE *fd, MalBlkPtr mb, MalStkPtr stk, int flg)
+{
+	int i,j;
+	InstrPtr p;
+	// Set the used bits properly
+	for(i=0; i< mb->vtop; i++)
+		clrVarUsed(mb,i);
+	for(i=0; i< mb->stop; i++){
+		p= getInstrPtr(mb,i);
+		for(j= p->retc; j<p->argc; j++)
+			setVarUsed(mb, getArg(p,j));
+		if( p->barrier)
+			for(j= 0; j< p->retc; j++)
+				setVarUsed(mb, getArg(p,j));
+	}
+	for (i = 0; i < mb->stop; i++)
+		fprintInstruction(fd, mb, stk, getInstrPtr(mb, i), flg);
+}
+
+/* initialize the static scope boundaries for all variables */
+void
+setVariableScope(MalBlkPtr mb)
 {
 	int pc, k, depth=0, dflow= -1;
 	InstrPtr p;
-	int *blk;
-	Lifespan span= newLifespan(mb);
-	str lang = putName("language"), dataflow= putName("dataflow");
 
-	if (span == NULL)
-		return NULL;
-
-	blk= (int *) GDKzalloc(sizeof(int)*mb->vtop);
-	if( blk == NULL){
-		GDKerror("setLifeSpan" MAL_MALLOC_FAIL);
-		GDKfree(span);
-		return NULL;
+	/* reset the scope admin */
+	for (k = 0; k < mb->vtop; k++)
+	if( isVarConstant(mb,k)){
+		setVarScope(mb,k,0);
+		setVarDeclared(mb,k,0);
+		setVarUpdated(mb,k,0);
+		setVarEolife(mb,k,mb->stop);
+	} else {
+		setVarScope(mb,k,0);
+		setVarDeclared(mb,k,0);
+		setVarUpdated(mb,k,0);
+		setVarEolife(mb,k,0);
 	}
 
 	for (pc = 0; pc < mb->stop; pc++) {
@@ -555,7 +606,7 @@ setLifespan(MalBlkPtr mb)
 			continue;
 
 		if( blockStart(p)){
-			if (getModuleId(p) == lang && getFunctionId(p) == dataflow){
+			if (getModuleId(p) && getFunctionId(p) && strcmp(getModuleId(p),"language")==0 && strcmp(getFunctionId(p),"dataflow")==0){
 				if( dflow != -1){
 					GDKerror("setLifeSpan nested dataflow blocks not allowed" );
 					mb->errors++;
@@ -567,18 +618,20 @@ setLifespan(MalBlkPtr mb)
 
 		for (k = 0; k < p->argc; k++) {
 			int v = getArg(p,k);
+			if( isVarConstant(mb,v) && getVarUpdated(mb,v) == 0)
+				setVarUpdated(mb,v, pc);
 
-			if (span[v].beginLifespan == 0 ){
-				span[v].beginLifespan = pc;
-				blk[v]= depth;
+			if ( getVarDeclared(mb,v) == 0 ){
+				setVarDeclared(mb,v, pc);
+				setVarScope(mb,v,depth);
 			}
 			if (k < p->retc )
-				span[v].lastUpdate= pc;
-			if ( blk[v] == depth )
-				span[v].endLifespan = pc;
+				setVarUpdated(mb,v, pc);
+			if ( getVarScope(mb,v) == depth )
+				setVarEolife(mb,v,pc);
 
-			if ( k >= p->retc && blk[v] < depth )
-				span[v].endLifespan = -1;	/* declared in outer scope*/
+			if ( k >= p->retc && getVarScope(mb,v) < depth )
+				setVarEolife(mb,v,-1);
 		}
 		/*
 		 * At a block exit we can finalize all variables defined within that block.
@@ -587,21 +640,23 @@ setLifespan(MalBlkPtr mb)
 		 */
 		if( blockExit(p) ){
 			for (k = 0; k < mb->vtop; k++)
-			if ( span[k].endLifespan == -1 )
-				span[k].endLifespan = pc;
-			else
-			if ( span[k].endLifespan == 0 && blk[k]==depth )
-				span[k].endLifespan = pc;
+			if ( getVarEolife(mb,k) == 0 && getVarScope(mb,k) ==depth )
+				setVarEolife(mb,k,pc);
+			else if ( getVarEolife(mb,k) == -1 )
+				setVarEolife(mb,k,pc);
+			
 			if( dflow == depth)
 				dflow= -1;
 			else depth--;
 		}
+		if( blockReturn(p)){
+			for (k = 0; k < p->argc; k++)
+				setVarEolife(mb,getArg(p,k),pc);
+		}
 	}
 	for (k = 0; k < mb->vtop; k++)
-	if ( span[k].endLifespan == 0 )
-		span[k].endLifespan = pc-2;/* generate them before the end */
-	GDKfree(blk);
-	return span;
+		if( getVarEolife(mb,k) == 0)
+			setVarEolife(mb,k, mb->stop-1);
 }
 
 int
@@ -669,18 +724,14 @@ void
 malGarbageCollector(MalBlkPtr mb)
 {
 	int i;
-	Lifespan span;
 
-	span = setLifespan(mb);
-	if ( span == NULL)
-		return ;
+	setVariableScope(mb);
 
 	for (i = 0; i < mb->vtop; i++)
-		if( isVarCleanup(mb,i) && getEndLifespan(span,i) >= 0) {
-			mb->var[i]->eolife = getEndLifespan(span,i);
-			mb->stmt[mb->var[i]->eolife]->gc |= GARBAGECONTROL;
+		if( isVarCleanup(mb,i) && getEndScope(mb,i) >= 0) {
+			setVarEolife(mb,i, getEndScope(mb,i));
+			mb->stmt[getVarEolife(mb,i)]->gc |= GARBAGECONTROL;
 		}
-	GDKfree(span);
 }
 /*
  * Variable declaration
@@ -714,22 +765,20 @@ void chkDeclarations(stream *out, MalBlkPtr mb){
 	int pc,i, k,l;
 	InstrPtr p;
 	short blks[MAXDEPTH], top= 0, blkId=1;
-	int *decl;
-	str lang = putName("language"), dataflow= putName("dataflow");
 	int dflow = -1;
 
-	decl = (int*) GDKzalloc(sizeof(int) * mb->vtop);
-	if ( decl == NULL) {
-		showScriptException(out, mb,0,SYNTAX, MAL_MALLOC_FAIL);
-		mb->errors = 1;
+	if( mb->errors)
 		return;
-	}
 	blks[top] = blkId;
 
+	/* initialize the scope */
+	for(i=0; i< mb->vtop; i++)
+		setVarScope(mb,i,0);
+
 	/* all signature variables are declared at outer level */
 	p= getInstrPtr(mb,0);
 	for(k=0;k<p->argc; k++)
-		decl[getArg(p,k)]= blkId;
+		setVarScope(mb, getArg(p,k), blkId);
 
 	for(pc=1;pc<mb->stop; pc++){
 		p= getInstrPtr(mb,pc);
@@ -739,7 +788,7 @@ void chkDeclarations(stream *out, MalBlkPtr mb){
 		for(k=p->retc;k<p->argc; k++) {
 			l=getArg(p,k);
 			setVarUsed(mb,l);
-			if( decl[l] == 0){
+			if( getVarScope(mb,l) == 0){
 				/*
 				 * The problem created here is that only variables are
 				 * recognized that are declared through instructions.
@@ -750,7 +799,7 @@ void chkDeclarations(stream *out, MalBlkPtr mb){
 				 * in the context of a global stack.
 				 */
 				if( p->barrier == CATCHsymbol){
-					decl[l] = blks[0];
+					setVarScope(mb, l, blks[0]);
 				} else
 				if( !( isVarConstant(mb, l) || isVarTypedef(mb,l)) &&
 					!isVarInit(mb,l) ) {
@@ -763,9 +812,9 @@ void chkDeclarations(stream *out, MalBlkPtr mb){
 			if( !isVarInit(mb,l) ){
 			    /* is the block still active ? */
 			    for( i=0; i<= top; i++)
-					if( blks[i] == decl[l] )
+					if( blks[i] == getVarScope(mb,l) )
 						break;
-			    if( i> top || blks[i]!= decl[l] ){
+			    if( i> top || blks[i]!= getVarScope(mb,l) ){
 			            showScriptException(out, mb,pc,TYPE,
 							"'%s' used outside scope",
 							getVarName(mb,l));
@@ -778,22 +827,21 @@ void chkDeclarations(stream *out, MalBlkPtr mb){
 		/* define variables */
 		for(k=0; k<p->retc; k++){
 			l= getArg(p,k);
-			if (isVarInit(mb, l) && decl[l] == 0) {
+			if (isVarInit(mb, l) && getVarScope(mb,l) == 0) {
 				/* first time we see this variable and it is already
 				 * initialized: assume it exists globally */
-				decl[l] = blks[0];
+				setVarScope(mb, l, blks[0]);
 			}
 			setVarInit(mb,l);
-			if( decl[l] == 0){
+			if( getVarScope(mb,l) == 0){
 				/* variable has not been defined yet */
 				/* exceptions are always declared at level 1 */
 				if( p->barrier == CATCHsymbol)
-					decl[l] = blks[0];
+					setVarScope(mb, l, blks[0]);
 				else
-					decl[l] = blks[top];
+					setVarScope(mb, l, blks[top]);
 #ifdef DEBUG_MAL_FCN
-				mnstr_printf(out,"defined %s in block %d\n",
-					getVarName(mb,l),decl[l]);
+				fprintf(stderr,"#defined %s in block %d\n", getVarName(mb,l), getVarScope(mb,l));
 #endif
 			}
 			if( blockCntrl(p) || blockStart(p) )
@@ -804,11 +852,10 @@ void chkDeclarations(stream *out, MalBlkPtr mb){
 				if( top == MAXDEPTH-2){
 					showScriptException(out, mb,pc,SYNTAX, "too deeply nested  MAL program");
 					mb->errors++;
-					GDKfree(decl);
 					return;
 				}
 				blkId++;
-				if (getModuleId(p) == lang && getFunctionId(p) == dataflow){
+				if (getModuleId(p) && getFunctionId(p) && strcmp(getModuleId(p),"language")==0 && strcmp(getFunctionId(p),"dataflow")== 0){
 					if( dflow != -1){
 						GDKerror("setLifeSpan nested dataflow blocks not allowed" );
 						mb->errors++;
@@ -817,12 +864,12 @@ void chkDeclarations(stream *out, MalBlkPtr mb){
 				} 
 				blks[++top]= blkId;
 #ifdef DEBUG_MAL_FCN
-				mnstr_printf(out,"new block %d at top %d\n",blks[top], top);
+				fprintf(stderr,"#new block %d at top %d\n",blks[top], top);
 #endif
 			}
 			if( blockExit(p) && top > 0 ){
 #ifdef DEBUG_MAL_FCN
-				mnstr_printf(out,"leave block %d at top %d\n",blks[top], top);
+				fprintf(stderr,"leave block %d at top %d\n",blks[top], top);
 #endif
 				if( dflow == blkId){
 					dflow = -1;
@@ -833,212 +880,12 @@ void chkDeclarations(stream *out, MalBlkPtr mb){
 				 * leading to uninitialized variables.
 				 */
 				for (l = 0; l < mb->vtop; l++)
-				if( decl[l] == blks[top]){
-					decl[l] =0;
+				if( getVarScope(mb,l) == blks[top]){
+					setVarScope(mb,l, 0);
 					clrVarInit(mb,l);
 				}
 			    top--;
 			}
 		}
 	}
-	GDKfree(decl);
 }
-
-/*
- * Data flow analysis.
- * Flow graph display is handy for debugging and analysis.
- * A better flow analysis is needed, which takes into account barrier blocks 
- */
-static void
-showOutFlow(MalBlkPtr mb, int pc, int varid, stream *f)
-{
-	InstrPtr p;
-	int i, k, found;
-
-	for (i = pc + 1; i < mb->stop - 1; i++) {
-		p = getInstrPtr(mb, i);
-		found = 0;
-		for (k = 0; k < p->argc; k++) {
-			if (p->argv[k] == varid) {
-				mnstr_printf(f, "n%d -> n%d\n", pc, i);
-				found++;
-			}
-		}
-		/* stop as soon you find a re-assignment */
-		for (k = 0; k < p->retc; k++) {
-			if (getArg(p, k) == varid)
-				i = mb->stop;
-		}
-		/* or a side-effect usage */
-		if (found &&
-			(p->retc == 0 || getArgType(mb, p, 0) == TYPE_void))
-			i = mb->stop;
-	}
-}
-
-static void
-showInFlow(MalBlkPtr mb, int pc, int varid, stream *f)
-{
-	InstrPtr p;
-	int i, k;
-
-	/* find last use, needed for operations with side effects */
-	for (i = pc - 1; i >= 0; i--) {
-		p = getInstrPtr(mb, i);
-		for (k = 0; k < p->argc; k++) {
-			if (p->argv[k] == varid ) {
-				mnstr_printf(f, "n%d -> n%d\n", i, pc);
-				return;
-			}
-		}
-	}
-}
-
-/*
- * We only display the minimal debugging information. The remainder
- * can be obtained through the profiler.
- */
-static void
-showFlowDetails(MalBlkPtr mb, MalStkPtr stk, InstrPtr p, int pc, stream *f)
-{
-	(void) mb;     /* fool the compiler */
-	(void) stk;     /* fool the compiler */
-	mnstr_printf(f, "n%d [fontsize=8, shape=box, label=\"%s\"]\n", pc, getFunctionId(p));
-}
-
-/* the stethoscope needs dot files for its graphical interface.
- * They are produced whenever a main() is called.
- * In all cases a single dot file is produced.
- */
-#define MAXFLOWGRAPHS 128
-
-int getFlowGraphs(MalBlkPtr mb, MalStkPtr stk, MalBlkPtr *mblist, MalStkPtr *stklist,int top);
-int getFlowGraphs(MalBlkPtr mb, MalStkPtr stk, MalBlkPtr *mblist, MalStkPtr *stklist,int top){
-	int i;
-	InstrPtr p;
-
-	for ( i=0; i<top; i++)
-	if ( mblist[i] == mb)
-		return top;
-
-	if ( top == MAXFLOWGRAPHS)
-		return top; /* just bail out */
-	mblist[top] = mb;
-	stklist[top++] = stk;
-	for (i=1; i < mb->stop; i++){
-		p = getInstrPtr(mb,i);
-		if ( p->token == FCNcall || p->token == FACcall )
-			top =getFlowGraphs(p->blk, 0,mblist, stklist, top);
-	}
-	return top;
-}
-
-void
-showFlowGraph(MalBlkPtr mb, MalStkPtr stk, str fname)
-{
-	stream *f;
-	InstrPtr p;
-	int i, j,k, stethoscope=0;
-	char mapimode = 0;
-	buffer *bufstr = NULL;
-	MalBlkPtr mblist[MAXFLOWGRAPHS];
-	MalStkPtr stklist[MAXFLOWGRAPHS];
-	int top =0;
-
-	(void) stk;     /* fool the compiler */
-
-	memset(mblist, 0, sizeof(mblist));
-	memset(stklist, 0, sizeof(stklist));
-
-	if (idcmp(fname, "stdout") == 0) {
-		f = GDKout;
-	} else if (idcmp(fname, "stdout-mapi") == 0) {
-		bufstr = buffer_create(8096);
-		f = buffer_wastream(bufstr, "bufstr_write");
-		mapimode = 1;
-	} else {
-		f = open_wastream(fname);
-	}
-	if ( f == NULL)
-		return;
-
-	top = getFlowGraphs(mb,stk,mblist,stklist,0);
-	if ( stethoscope == 0)
-		top =1;
-	for( j=0; j< top; j++){
-		mb = mblist[j];
-		stk = stklist[j];
-		if (mb == 0 || (mb->dotfile && stethoscope))
-			continue; /* already sent */
-		p = getInstrPtr(mb, 0);
-		mnstr_printf(f, "digraph %s {\n", getFunctionId(p));
-		p = getInstrPtr(mb, 0);
-		showFlowDetails(mb, stk, p, 0, f);
-		for (k = p->retc; k < p->argc; k++) {
-			showOutFlow(mb, 0, p->argv[k], f);
-		}
-		for (i = 1; i < mb->stop; i++) {
-			p = getInstrPtr(mb, i);
-
-			showFlowDetails(mb, stk, p, i, f);
-
-			for (k = 0; k < p->retc; k++)
-				showOutFlow(mb, i, p->argv[k], f);
-
-			if (p->retc == 0 || getArgType(mb, p, 0) == TYPE_void) /* assume side effects */
-				for (k = p->retc; k < p->argc; k++)
-					if (getArgType(mb, p, k) != TYPE_void &&
-						!isVarConstant(mb, getArg(p, k)))
-						showOutFlow(mb, i, p->argv[k], f);
-
-			if (getFunctionId(p) == 0)
-				for (k = 0; k < p->retc; k++)
-					if (getArgType(mb, p, k) != TYPE_void)
-						showInFlow(mb, i, p->argv[k], f);
-			if (p->token == ENDsymbol)
-				break;
-		}
-		mnstr_printf(f, "}\n");
-		mb->dotfile++;
-	}
-
-	if (mapimode == 1) {
-		size_t maxlen = 0;
-		size_t rows = 0;
-		str buf = buffer_get_buf(bufstr);
-		str line, oline;
-
-		/* calculate width of column, and the number of tuples */
-		oline = buf;
-		while ((line = strchr(oline, '\n')) != NULL) {
-			if ((size_t) (line - oline) > maxlen)
-				maxlen = line - oline;
-			rows++;
-			oline = line + 1;
-		} /* see printf before this mapimode if, last line ends with \n */
-
-		/* write mapi header */
-		if ( f == GDKout) {
-			mnstr_printf(f, "&1 0 " SZFMT " 1 " SZFMT "\n",
-					/* type id rows columns tuples */ rows, rows);
-			mnstr_printf(f, "%% .dot # table_name\n");
-			mnstr_printf(f, "%% dot # name\n");
-			mnstr_printf(f, "%% clob # type\n");
-			mnstr_printf(f, "%% " SZFMT " # length\n", maxlen);
-		}
-		oline = buf;
-		while ((line = strchr(oline, '\n')) != NULL) {
-			*line++ = '\0';
-			mnstr_printf(GDKout, "=%s\n", oline);
-			oline = line;
-		}
-		free(buf);
-	}
-	if (f != GDKout) {
-		if (!stethoscope ) {
-			MT_sleep_ms(4000); /* delay for stethoscope */
-			close_stream(f);
-		}
-	}
-}
-
