@@ -3,7 +3,7 @@
  * License, v. 2.0.  If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
- * Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.
+ * Copyright 1997 - July 2008 CWI, August 2008 - 2017 MonetDB B.V.
  */
 
 /*
@@ -19,6 +19,7 @@
 #include "mal_resolve.h"
 #include "mal_namespace.h"
 #include "mal_private.h"
+#include "mal_linker.h"
 
 static malType getPolyType(malType t, int *polytype);
 static int updateTypeMap(int formal, int actual, int polytype[MAXTYPEVAR]);
@@ -88,9 +89,8 @@ int polyVector[MAXTYPEVAR];
  */
 #define prepostProcess(tp, p, b, mb)					\
 	do {												\
-		if (findGDKtype(tp) == TYPE_bat ||				\
-			isaBatType(tp) ||							\
-			findGDKtype(tp) == TYPE_str ||				\
+		if( isaBatType(tp) ||							\
+			ATOMtype(tp) == TYPE_str ||				\
 			(!isPolyType(tp) && tp < TYPE_any &&		\
 			 tp >= 0 && ATOMextern(tp))) {				\
 			getInstrPtr(mb, 0)->gc |= GARBAGECONTROL;	\
@@ -110,7 +110,7 @@ findFunctionType(stream *out, Module scope, MalBlkPtr mb, InstrPtr p, int silent
 	int returns[256];
 	int *returntype = NULL;
 	/*
-	 * Within a module find the subscope to locate the element in its list
+	 * Within a module find the element in its list
 	 * of symbols. A skiplist is used to speed up the search for the
 	 * definition of the function.
 	 *
@@ -126,17 +126,19 @@ findFunctionType(stream *out, Module scope, MalBlkPtr mb, InstrPtr p, int silent
 	 * Simplify polytype using a map into the concrete argument table.
 	 */
 	m = scope;
-	s = m->subscope[(int) (getSubScope(getFunctionId(p)))];
+	s = m->space[(int) (getSymbolIndex(getFunctionId(p)))];
 	if (s == 0)
 		return -1;
 
 	if ( p->retc < 256){
-		for(i=0; i< p->retc; i++) returns[i] = 0;
+		for (i=0; i< p->retc; i++)
+			returns[i] = 0;
 		returntype = returns;
-	} else 
+	} else {
 		returntype = (int *) GDKzalloc(p->retc * sizeof(int));
-	if (returntype == 0)
-		return -1;
+		if (returntype == 0)
+			return -1;
+	}
 
 	while (s != NULL) {			/* single scope element check */
 		if (getFunctionId(p) != s->name) {
@@ -166,11 +168,11 @@ findFunctionType(stream *out, Module scope, MalBlkPtr mb, InstrPtr p, int silent
 
 #ifdef DEBUG_MAL_RESOLVE
 		if (tracefcn) {
-			mnstr_printf(out, "-->resolving\n");
-			printInstruction(out, mb, 0, p, LIST_MAL_ALL);
-			mnstr_printf(out, "++> test against signature\n");
-			printInstruction(out, s->def, 0, getSignature(s), LIST_MAL_ALL);
-			mnstr_printf(out, " %s \n", sig->polymorphic ? "polymorphic" : "");
+			fprintf(stderr, "-->resolving\n");
+			fprintInstruction(stderr, mb, 0, p, LIST_MAL_ALL);
+			fprintf(stderr, "++> test against signature\n");
+			fprintInstruction(stderr, s->def, 0, getSignature(s), LIST_MAL_ALL);
+			fprintf(stderr, " %s \n", sig->polymorphic ? "polymorphic" : "");
 		}
 #endif
 		/*
@@ -289,7 +291,7 @@ findFunctionType(stream *out, Module scope, MalBlkPtr mb, InstrPtr p, int silent
 #ifdef DEBUG_MAL_RESOLVE
 					char *ftpe = getTypeName(formal);
 					char *atpe = getTypeName(actual);
-					mnstr_printf(out, "unmatched %d formal %s actual %s\n",
+					fprintf(stderr, "unmatched %d formal %s actual %s\n",
 								 i, ftpe, atpe);
 					GDKfree(ftpe);
 					GDKfree(atpe);
@@ -310,7 +312,7 @@ findFunctionType(stream *out, Module scope, MalBlkPtr mb, InstrPtr p, int silent
 #ifdef DEBUG_MAL_RESOLVE
 		if (tracefcn) {
 			char *tpe, *tpe2;
-			mnstr_printf(out, "finished %s.%s unmatched=%d polymorphic=%d %d\n",
+			fprintf(stderr, "finished %s.%s unmatched=%d polymorphic=%d %d\n",
 						 getModuleId(sig), getFunctionId(sig), unmatched,
 						 sig->polymorphic, p == sig);
 			if (sig->polymorphic) {
@@ -318,17 +320,17 @@ findFunctionType(stream *out, Module scope, MalBlkPtr mb, InstrPtr p, int silent
 				for (l = 0; l < 2 * p->argc; l++)
 					if (polytype[l] != TYPE_any) {
 						tpe = getTypeName(polytype[l]);
-						mnstr_printf(out, "poly %d %s\n", l, tpe);
+						fprintf(stderr, "poly %d %s\n", l, tpe);
 						GDKfree(tpe);
 					}
 			}
-			mnstr_printf(out, "-->resolving\n");
-			printInstruction(out, mb, 0, p, LIST_MAL_ALL);
-			mnstr_printf(out, "++> test against signature\n");
-			printInstruction(out, s->def, 0, getSignature(s), LIST_MAL_ALL);
+			fprintf(stderr, "-->resolving\n");
+			fprintInstruction(stderr, mb, 0, p, LIST_MAL_ALL);
+			fprintf(stderr, "++> test against signature\n");
+			fprintInstruction(stderr, s->def, 0, getSignature(s), LIST_MAL_ALL);
 			tpe = getTypeName(getArgType(mb, p, unmatched));
 			tpe2 = getTypeName(getArgType(s->def, sig, unmatched));
-			mnstr_printf(out, "\nmismatch unmatched %d test %s poly %s\n",
+			fprintf(stderr, "\nmismatch unmatched %d test %s poly %s\n",
 						 unmatched, tpe, tpe2);
 			GDKfree(tpe);
 			GDKfree(tpe2);
@@ -418,7 +420,7 @@ findFunctionType(stream *out, Module scope, MalBlkPtr mb, InstrPtr p, int silent
 		 * to be garbage collected.
 		 */
 		for (i = p->retc; i < p->argc; i++)
-			if (findGDKtype(getArgType(mb, p, i)) == TYPE_str ||
+			if (ATOMtype(getArgType(mb, p, i)) == TYPE_str ||
 				getArgType(mb, p, i) == TYPE_bat ||
 				isaBatType(getArgType(mb, p, i)) ||
 				(!isPolyType(getArgType(mb, p, i)) &&
@@ -459,11 +461,11 @@ findFunctionType(stream *out, Module scope, MalBlkPtr mb, InstrPtr p, int silent
 
 #ifdef DEBUG_MAL_RESOLVE
 		if (tracefcn) {
-			printInstruction(out, mb, 0, p, LIST_MAL_ALL);
-			mnstr_printf(out, "Finished matching\n");
+			fprintInstruction(stderr, mb, 0, p, LIST_MAL_ALL);
+			fprintf(stderr, "Finished matching\n");
 		}
 #endif
-		if (returntype && returntype != returns)
+		if (returntype != returns)
 			GDKfree(returntype);
 		return s1;
 	} /* while */
@@ -473,7 +475,7 @@ findFunctionType(stream *out, Module scope, MalBlkPtr mb, InstrPtr p, int silent
 	 * arguments, but that clashes with one of the target variables.
 	 */
   wrapup:
-	if (returntype && returntype != returns)
+	if (returntype != returns)
 		GDKfree(returntype);
 	return -3;
 }
@@ -485,7 +487,7 @@ resolveType(int dsttype, int srctype)
 	if (tracefcn) {
 		char *dtpe = getTypeName(dsttype);
 		char *stpe = getTypeName(srctype);
-		mnstr_printf(GDKout, "resolveType dst %s (%d) %s(%d)\n",
+		fprintf(stderr, "resolveType dst %s (%d) %s(%d)\n",
 					 dtpe, dsttype, stpe, srctype);
 		GDKfree(dtpe);
 		GDKfree(stpe);
@@ -506,8 +508,8 @@ resolveType(int dsttype, int srctype)
 		return dsttype;
 	if (isaBatType(dsttype) && isaBatType(srctype)) {
 		int t1, t2, t3;
-		t1 = getColumnType(dsttype);
-		t2 = getColumnType(srctype);
+		t1 = getBatType(dsttype);
+		t2 = getBatType(srctype);
 		if (t1 == t2)
 			t3 = t1;
 		else if (t1 == TYPE_any)
@@ -517,29 +519,29 @@ resolveType(int dsttype, int srctype)
 		else {
 #ifdef DEBUG_MAL_RESOLVE
 			if (tracefcn)
-				mnstr_printf(GDKout, "Tail can not be resolved \n");
+				fprintf(stderr, "Tail can not be resolved \n");
 #endif
 			return -1;
 		}
 #ifdef DEBUG_MAL_RESOLVE
 		if (tracefcn) {
-			int i2 = getColumnIndex(dsttype);
+			int i2 = getTypeIndex(dsttype);
 			char *tpe1, *tpe2, *tpe3; 
 			tpe1 = getTypeName(t1);
 			tpe2 = getTypeName(t2);
 			tpe3 = getTypeName(t3);
-			mnstr_printf(GDKout, "resolved to bat[:oid,:%s] bat[:oid,:%s]->bat[:oid,%s:%d]\n",
+			fprintf(stderr, "resolved to bat[:oid,:%s] bat[:oid,:%s]->bat[:oid,%s:%d]\n",
 						 tpe1, tpe2, tpe3, i2);
 			GDKfree(tpe1);
 			GDKfree(tpe2);
 			GDKfree(tpe3);
 		}
 #endif
-		return newBatType(TYPE_void, t3);
+		return newBatType(t3);
 	}
 #ifdef DEBUG_MAL_RESOLVE
 	if (tracefcn)
-		mnstr_printf(GDKout, "Can not be resolved \n");
+		fprintf(stderr, "Can not be resolved \n");
 #endif
 	return -1;
 }
@@ -587,7 +589,7 @@ typeChecker(stream *out, Module scope, MalBlkPtr mb, InstrPtr p, int silent)
 
 	p->typechk = TYPE_UNKNOWN;
 	olderrors = mb->errors;
-	if (p->fcn && p->token >= FCNcall && p->token <= PATcall) {
+	if ((p->fcn || p->blk) && p->token >= FCNcall && p->token <= PATcall) {
 		p->token = ASSIGNsymbol;
 		p->fcn = NULL;
 		p->blk = NULL;
@@ -627,15 +629,19 @@ typeChecker(stream *out, Module scope, MalBlkPtr mb, InstrPtr p, int silent)
 		if (!isaSignature(p) && !getInstrPtr(mb, 0)->polymorphic) {
 			mb->errors++;
 			if (!silent) {
-				char *errsig;
+				if (!malLibraryEnabled(p->modname)) {
+					dumpExceptionsToStream(out, malLibraryHowToEnable(p->modname));
+				} else {
+					char *errsig;
 
-				errsig = instruction2str(mb,0,p,(LIST_MAL_NAME | LIST_MAL_TYPE | LIST_MAL_VALUE));
-				showScriptException(out, mb, getPC(mb, p), TYPE,
-									"'%s%s%s' undefined in: %s",
-									(getModuleId(p) ? getModuleId(p) : ""),
-									(getModuleId(p) ? "." : ""),
-									getFunctionId(p), errsig);
-				GDKfree(errsig);
+					errsig = instruction2str(mb,0,p,(LIST_MAL_NAME | LIST_MAL_TYPE | LIST_MAL_VALUE));
+					showScriptException(out, mb, getPC(mb, p), TYPE,
+										"'%s%s%s' undefined in: %s",
+										(getModuleId(p) ? getModuleId(p) : ""),
+										(getModuleId(p) ? "." : ""),
+										getFunctionId(p), errsig?errsig:"failed instruction2str()");
+					GDKfree(errsig);
+				}
 			} else
 				mb->errors = olderrors;
 			p->typechk = TYPE_UNKNOWN;
@@ -696,8 +702,8 @@ typeChecker(stream *out, Module scope, MalBlkPtr mb, InstrPtr p, int silent)
 	if (p->barrier && p->retc == p->argc)
 		for (k = 0; k < p->retc; k++) {
 			int tpe = getArgType(mb, p, k);
-			if (findGDKtype(tpe) == TYPE_bat ||
-				findGDKtype(tpe) == TYPE_str ||
+			if (isaBatType(tpe)  ||
+				ATOMtype(tpe) == TYPE_str ||
 				(!isPolyType(tpe) && tpe < MAXATOMS && ATOMextern(tpe)))
 				setVarCleanup(mb, getArg(p, k));
 		}
@@ -728,7 +734,8 @@ chkTypes(stream *out, Module s, MalBlkPtr mb, int silent)
 	for (i = 0; i < mb->stop; i++) {
 		p = getInstrPtr(mb, i);
 		assert (p != NULL);
-		typeChecker(out, s, mb, p, silent);
+		if (p->typechk != TYPE_RESOLVED)
+			typeChecker(out, s, mb, p, silent);
 		if (mb->errors)
 			return;
 
@@ -749,6 +756,8 @@ chkInstruction(stream *out, Module s, MalBlkPtr mb, InstrPtr p)
 {
 	int olderrors= mb->errors;
 	int error;
+
+	p->typechk = TYPE_UNKNOWN;
 	typeChecker(out, s, mb, p, TRUE);
 	error = mb->errors;
 	mb->errors = olderrors;
@@ -801,13 +810,13 @@ getPolyType(malType t, int *polytype)
 	int ti;
 	int tail;
 
-	ti = getColumnIndex(t);
+	ti = getTypeIndex(t);
 	if (!isaBatType(t) && ti > 0)
 		return polytype[ti];
 
-	tail = ti == 0 ? getColumnType(t) : polytype[ti];
+	tail = ti == 0 ? getBatType(t) : polytype[ti];
 	if (isaBatType(t)) 
-		return newBatType(TYPE_void, tail);
+		return newBatType(tail);
 	return tail;
 }
 
@@ -830,20 +839,20 @@ updateTypeMap(int formal, int actual, int polytype[MAXTYPEVAR])
 #ifdef DEBUG_MAL_RESOLVE
 	{
 		char *tpe1 = getTypeName(formal), *tpe2 = getTypeName(actual);
-		mnstr_printf(GDKout, "updateTypeMap:formal %s actual %s\n", tpe1, tpe2);
+		fprintf(stderr, "updateTypeMap:formal %s actual %s\n", tpe1, tpe2);
 		GDKfree(tpe1);
 		GDKfree(tpe2);
 	}
 #endif
 
-	if ((h = getColumnIndex(formal))) {
+	if ((h = getTypeIndex(formal))) {
 		if (isaBatType(actual) && !isaBatType(formal) &&
 			(polytype[h] == TYPE_any || polytype[h] == actual)) {
 			polytype[h] = actual;
 			ret = 0;
 			goto updLabel;
 		}
-		t = getColumnType(actual);
+		t = getBatType(actual);
 		if (t != polytype[h]) {
 			if (polytype[h] == TYPE_bat && isaBatType(actual))
 				ret = 0;
@@ -861,7 +870,7 @@ updateTypeMap(int formal, int actual, int polytype[MAXTYPEVAR])
 	}
   updLabel:
 #ifdef DEBUG_MAL_RESOLVE
-	mnstr_printf(GDKout, "updateTypeMap returns: %d\n", ret);
+	fprintf(stderr, "updateTypeMap returns: %d\n", ret);
 #endif
 	return ret;
 }
