@@ -3,7 +3,7 @@
  * License, v. 2.0.  If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
- * Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.
+ * Copyright 1997 - July 2008 CWI, August 2008 - 2017 MonetDB B.V.
  */
 
 #ifndef _MAL_RESOURCE_H
@@ -18,7 +18,7 @@
 
 //#define heapinfo(X,Id)	(((X) && (X)->base && ((X)->parentid == 0 || (X)->parentid == Id)) ? (X)->free : 0)
 #define heapinfo(X,Id)	(((X) && (X)->base ) ? (X)->free : 0)
-#define hashinfo(X,Id) ( (X)? heapinfo((X)->heap, Id):0)
+#define hashinfo(X,Id) ((X) && (X) != (Hash *) 1 ? heapinfo((X)->heap, Id) : 0)
 
 #define USE_MAL_ADMISSION
 #ifdef USE_MAL_ADMISSION
