@@ -3,7 +3,7 @@
  * License, v. 2.0.  If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
- * Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.
+ * Copyright 1997 - July 2008 CWI, August 2008 - 2017 MonetDB B.V.
  */
 
 /* (author) M.L. Kersten
@@ -12,19 +12,13 @@
 #include "mal_session.h"
 #include "mal_instruction.h" /* for pushEndInstruction() */
 #include "mal_interpreter.h" /* for showErrors(), runMAL(), garbageElement() */
-#include "mal_linker.h"	     /* for initLibraries() */
 #include "mal_parser.h"	     /* for parseMAL() */
 #include "mal_namespace.h"
-#include "mal_readline.h"
+#include "mal_builder.h"
 #include "mal_authorize.h"
 #include "mal_private.h"
 #include <gdk.h>	/* for opendir and friends */
 
-#ifdef HAVE_EMBEDDED
-// FIXME:
-//#include "mal_init_inline.h"
-#endif
-
 /*
  * The MonetDB server uses a startup script to boot the system.
  * This script is an ordinary MAL program, but will mostly
@@ -38,29 +32,43 @@ malBootstrap(void)
 	str msg, bootfile = "mal_init", s = NULL;
 
 	c = MCinitClient((oid) 0, 0, 0);
-	assert(c != NULL);
+	if (!c) {
+		GDKerror("malBootstrap: Failed to initialize client");
+		return 0;
+	}
 	c->nspace = newModule(NULL, putName("user"));
-	initLibraries();
 	if ( (msg = defaultScenario(c)) ) {
-		GDKfree(msg);
-		GDKerror("Failed to initialise default scenario");
+		freeException(msg);
+		MALexitClient(c);
+		GDKerror("malBootstrap: Failed to initialize default scenario");
 		return 0;
 	}
 	MSinitClientPrg(c, "user", "main");
-	(void) MCinitClientThread(c);
+	if( MCinitClientThread(c) < 0){
+		MALexitClient(c);
+		GDKerror("malBootstrap: Failed to create client thread");
+		return 0;
+	}
 	s = malInclude(c, bootfile, 0);
 	if (s != NULL) {
-		mnstr_printf(GDKout, "!%s\n", s);
+		GDKerror("malBootstrap: Failed to load startup script %s", s);
+		MALexitClient(c);
 		GDKfree(s);
 		return 0;
 	}
 	pushEndInstruction(c->curprg->def);
 	chkProgram(c->fdout, c->nspace, c->curprg->def);
-	if (c->curprg->def->errors)
-		showErrors(c);
+	if (c->curprg->def->errors) {
+		GDKerror("malBootstrap: Failed to check startup script %s", GDKerrbuf);
+		MALexitClient(c);
+		return 0;
+	}
 	s = MALengine(c);
-	if (s)
+	if (s != MAL_SUCCEED) {
 		GDKfree(s);
+		return 0;
+	}
+	MALexitClient(c);
 	return 1;
 }
 
@@ -95,6 +103,8 @@ MSresetClientPrg(Client cntxt)
 	p->gc = 0;
 	p->retc = 1;
 	p->argc = 1;
+	setModuleId(p, putName("user"));
+	setFunctionId(p, putName("main"));
 	/* remove any MAL history */
 	if (mb->history) {
 		freeMalBlk(mb->history);
@@ -113,6 +123,10 @@ MSinitClientPrg(Client cntxt, str mod, str nme)
 		return;
 	}
 	cntxt->curprg = newFunction(putName("user"), putName(nme), FUNCTIONsymbol);
+	if( cntxt->curprg == 0){
+		GDKerror("MSinitClientPrg" "Failed to create function");
+		return;
+	}
 	mb = cntxt->curprg->def;
 	p = getSignature(cntxt->curprg);
 	if (mod)
@@ -138,17 +152,16 @@ exit_streams( bstream *fin, stream *fout )
 {
 	if (fout && fout != GDKstdout) {
 		mnstr_flush(fout);
-		mnstr_close(fout);
-		mnstr_destroy(fout);
+		close_stream(fout);
 	}
-	if (fin) 
-		(void) bstream_destroy(fin);
+	if (fin)
+		bstream_destroy(fin);
 }
 
 const char* mal_enableflag = "mal_for_all";
 
 void
-MSscheduleClient(str command, str challenge, bstream *fin, stream *fout)
+MSscheduleClient(str command, str challenge, bstream *fin, stream *fout, protocol_version protocol, size_t blocksize, int compute_column_widths)
 {
 	char *user = command, *algo = NULL, *passwd = NULL, *lang = NULL;
 	char *database = NULL, *s, *dbname;
@@ -241,11 +254,11 @@ MSscheduleClient(str command, str challenge, bstream *fin, stream *fout)
 		/* access control: verify the credentials supplied by the user,
 		 * no need to check for database stuff, because that is done per
 		 * database itself (one gets a redirect) */
-		err = AUTHcheckCredentials(&uid, root, &user, &passwd, &challenge, &algo);
+		err = AUTHcheckCredentials(&uid, root, user, passwd, challenge, algo);
 		if (err != MAL_SUCCEED) {
 			mnstr_printf(fout, "!%s\n", err);
 			exit_streams(fin, fout);
-			GDKfree(err);
+			freeException(err);
 			GDKfree(command);
 			return;
 		}
@@ -264,7 +277,6 @@ MSscheduleClient(str command, str challenge, bstream *fin, stream *fout)
 		/* move this back !! */
 		if (c->nspace == 0) {
 			c->nspace = newModule(NULL, putName("user"));
-			c->nspace->outer = mal_clients[0].nspace->outer;
 		}
 
 		if ((s = setScenario(c, lang)) != NULL) {
@@ -301,6 +313,11 @@ MSscheduleClient(str command, str challenge, bstream *fin, stream *fout)
 	 * demand. */
 
 	/* fork a new thread to handle this client */
+
+	c->protocol = protocol;
+	c->blocksize = blocksize;
+	c->compute_column_widths = compute_column_widths;
+
 	mnstr_settimeout(c->fdin->s, 50, GDKexiting);
 	MSserveClient(c);
 }
@@ -348,21 +365,16 @@ void
 MSresetVariables(Client cntxt, MalBlkPtr mb, MalStkPtr glb, int start)
 {
 	int i;
-	bit *used = GDKzalloc(mb->vtop * sizeof(bit));
-	if( used == NULL){
-		GDKerror("MSresetVariables" MAL_MALLOC_FAIL);
-		return;
-	}
 
 	for (i = 0; i < start && start < mb->vtop; i++)
-		used[i] = 1;
+		setVarUsed(mb,i);
 	if (mb->errors == 0)
 		for (i = start; i < mb->vtop; i++) {
-			if (used[i] || !isTmpVar(mb, i)) {
-				assert(!mb->var[i]->value.vtype || isVarConstant(mb, i));
-				used[i] = 1;
+			if (isVarUsed(mb,i) || !isTmpVar(mb,i)){
+				assert(!mb->var[i].value.vtype || isVarConstant(mb, i));
+				setVarUsed(mb,i);
 			}
-			if (glb && !used[i]) {
+			if (glb && !isVarUsed(mb,i)) {
 				if (isVarConstant(mb, i))
 					garbageElement(cntxt, &glb->stk[i]);
 				/* clean stack entry */
@@ -373,8 +385,7 @@ MSresetVariables(Client cntxt, MalBlkPtr mb, MalStkPtr glb, int start)
 		}
 
 	if (mb->errors == 0)
-		trimMalVariables_(mb, used, glb);
-	GDKfree(used);
+		trimMalVariables_(mb, glb);
 }
 
 /*
@@ -402,7 +413,7 @@ MSserveClient(void *dummy)
 		c->glb = newGlobalStack(MAXGLOBALS + mb->vsize);
 	if (c->glb == NULL) {
 		showException(c->fdout, MAL, "serveClient", MAL_MALLOC_FAIL);
-		c->mode = FINISHCLIENT + 1; /* == RUNCLIENT */
+		c->mode = RUNCLIENT;
 	} else {
 		c->glb->stktop = mb->vtop;
 		c->glb->blk = mb;
@@ -412,12 +423,13 @@ MSserveClient(void *dummy)
 		msg = defaultScenario(c);
 	if (msg) {
 		showException(c->fdout, MAL, "serveClient", "could not initialize default scenario");
-		c->mode = FINISHCLIENT + 1; /* == RUNCLIENT */
-		GDKfree(msg);
+		c->mode = RUNCLIENT;
+		freeException(msg);
 	} else {
 		do {
 			do {
-				runScenario(c);
+				msg = runScenario(c);
+				freeException(msg);
 				if (c->mode == FINISHCLIENT)
 					break;
 				resetScenario(c);
@@ -431,8 +443,19 @@ MSserveClient(void *dummy)
 	/*
 	 * At this stage we should clean out the MAL block
 	 */
-	freeMalBlk(c->curprg->def);
-	c->curprg->def = 0;
+	if (c->backup) {
+		assert(0);
+		freeSymbol(c->backup);
+		c->backup = 0;
+	}
+	if (c->curprg) {
+		assert(0);
+		freeSymbol(c->curprg);
+		c->curprg = 0;
+	}
+	if (c->nspace) {
+		assert(0);
+	}
 
 	if (c->mode > FINISHCLIENT) {
 		if (isAdministrator(c) /* && moreClients(0)==0 */) {
@@ -443,6 +466,11 @@ MSserveClient(void *dummy)
 	}
 	if (!isAdministrator(c))
 		MCcloseClient(c);
+	if (c->nspace && strcmp(c->nspace->name, "user") == 0) {
+		GDKfree(c->nspace->space);
+		GDKfree(c->nspace);
+		c->nspace = NULL;
+	}
 }
 
 /*
@@ -472,22 +500,23 @@ MALexitClient(Client c)
 	if (c->bak)
 		return NULL;
 	c->mode = FINISHCLIENT;
+	if (c->backup) {
+		assert(0);
+		freeSymbol(c->backup);
+		c->backup = NULL;
+	}
+	/* should be in the nspace */
+	c->curprg = NULL;
+	if (c->nspace) {
+		freeModule(c->nspace);
+		c->nspace = NULL;
+	}
 	return NULL;
 }
 
 str
 MALreader(Client c)
 {
-	int r = 1;
-#ifndef HAVE_EMBEDDED
-	if (c == mal_clients) {
-		r = readConsole(c);
-		if (r < 0 && c->fdin->eof == 0)
-			r = MCreadClient(c);
-		if (r > 0)
-			return MAL_SUCCEED;
-	} else
-#endif
 	if (MCreadClient(c) > 0)
 		return MAL_SUCCEED;
 	MT_lock_set(&mal_contextLock);
@@ -509,7 +538,9 @@ MALparser(Client c)
 	c->curprg->def->errors = 0;
 	oldstate = *c->curprg->def;
 
-	prepareMalBlk(c->curprg->def, CURRENT(c));
+	if( prepareMalBlk(c->curprg->def, CURRENT(c))){
+		throw(MAL, "mal.parser", MAL_MALLOC_FAIL);
+	}
 	if (parseMAL(c, c->curprg, 0) || c->curprg->def->errors) {
 		/* just complete it for visibility */
 		pushEndInstruction(c->curprg->def);
