@@ -3,7 +3,7 @@
  * License, v. 2.0.  If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
- * Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.
+ * Copyright 1997 - July 2008 CWI, August 2008 - 2017 MonetDB B.V.
  */
 
 #ifndef _MAL_STACK_H_
@@ -11,9 +11,6 @@
 #include "mal.h"
 
 #define stackSize(CNT) (sizeof(ValRecord)*(CNT) + offsetof(MalStack, stk))
-#define newStack(S,CNT) S= (MalStkPtr) GDKzalloc(stackSize(CNT));\
-		(S)->stksize=CNT;
-
 
 mal_export MalStkPtr newGlobalStack(int size);
 mal_export MalStkPtr reallocGlobalStack(MalStkPtr s, int cnt);
