@@ -3,7 +3,7 @@
  * License, v. 2.0.  If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
- * Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.
+ * Copyright 1997 - July 2008 CWI, August 2008 - 2017 MonetDB B.V.
  */
 
 #ifndef _OPT_DATAFLOW_
@@ -11,8 +11,6 @@
 #include "opt_prelude.h"
 #include "opt_support.h"
 
-#define OPTDEBUGdataflow  if ( optDebug & ((lng) 1 <<DEBUG_OPT_DATAFLOW) )
-opt_export void removeDataflow(MalBlkPtr mb);
-opt_export int OPTdataflowImplementation(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr p);
+mal_export str OPTdataflowImplementation(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr p);
 
 #endif
