@@ -3,7 +3,7 @@
  * License, v. 2.0.  If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
- * Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.
+ * Copyright 1997 - July 2008 CWI, August 2008 - 2017 MonetDB B.V.
  */
 
 /*
@@ -17,9 +17,10 @@
  */
 #include "monetdb_config.h"
 #include "sql_user.h"
+
+#include "bat5.h"
 #include "sql_mvc.h"
 #include "sql_privileges.h"
-#include "bat5.h"
 #include "mal_interpreter.h"
 #include "mal_authorize.h"
 #include "mcrypt.h"
@@ -60,7 +61,7 @@ monet5_drop_user(ptr _mvc, str user)
 	str err;
 	Client c = MCgetClient(m->clientid);
 
-	err = AUTHremoveUser(c, &user);
+	err = AUTHremoveUser(c, user);
 	if (err !=MAL_SUCCEED) {
 		(void) sql_error(m, 02, "DROP USER: %s", getExceptionMessage(err));
 		_DELETE(err);
@@ -105,7 +106,7 @@ monet5_create_user(ptr _mvc, str user, str passwd, char enc, str fullname, sqlid
 		pwd = passwd;
 	}
 	/* add the user to the M5 authorisation administration */
-	ret = AUTHaddUser(&uid, c, &user, &pwd);
+	ret = AUTHaddUser(&uid, c, user, pwd);
 	if (!enc)
 		free(pwd);
 	if (ret != MAL_SUCCEED)
@@ -128,8 +129,10 @@ monet5_find_user(ptr mp, str user)
 	Client c = MCgetClient(m->clientid);
 	str err;
 
-	if ((err = AUTHgetUsers(&uid, &nme, c)) != MAL_SUCCEED)
+	if ((err = AUTHgetUsers(&uid, &nme, c)) != MAL_SUCCEED) {
+		_DELETE(err);
 		return -1;
+	}
 	p = BUNfnd(nme, user);
 	BBPunfix(uid->batCacheid);
 	BBPunfix(nme->batCacheid);
@@ -157,13 +160,11 @@ db_users_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
 str
 db_password_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
 {
-	str ret = NULL;
 	str *hash = getArgReference_str(stk, pci, 0);
 	str *user = getArgReference_str(stk, pci, 1);
 	(void) mb;
 
-	ret = AUTHgetPasswordHash(hash, cntxt, user);
-	return ret;
+	return AUTHgetPasswordHash(hash, cntxt, *user);
 }
 
 static void
@@ -193,7 +194,13 @@ monet5_create_privileges(ptr _mvc, sql_schema *s)
 	   with the approriate scenario (sql) */
 	mvc_create_func(m, NULL, s, "db_users", ops, res, F_UNION, FUNC_LANG_SQL, "sql", "db_users", "CREATE FUNCTION db_users () RETURNS TABLE( name varchar(2048)) EXTERNAL NAME sql.db_users;", FALSE, FALSE);
 
-	t = mvc_create_view(m, s, "users", SQL_PERSIST, "SELECT u.\"name\" AS \"name\", " "ui.\"fullname\", ui.\"default_schema\" " "FROM db_users() AS u LEFT JOIN " "\"sys\".\"db_user_info\" AS ui " "ON u.\"name\" = ui.\"name\" " ";", 1);
+	t = mvc_create_view(m, s, "users", SQL_PERSIST,
+			    "SELECT u.\"name\" AS \"name\", "
+			    "ui.\"fullname\", ui.\"default_schema\" "
+			    "FROM db_users() AS u LEFT JOIN "
+			    "\"sys\".\"db_user_info\" AS ui "
+			    "ON u.\"name\" = ui.\"name\" "
+			    ";", 1);
 	mvc_create_column_(m, t, "name", "varchar", 1024);
 	mvc_create_column_(m, t, "fullname", "varchar", 2024);
 	mvc_create_column_(m, t, "default_schema", "int", 9);
@@ -249,44 +256,49 @@ monet5_alter_user(ptr _mvc, str user, str passwd, char enc, sqlid schema_id, str
 			opwd = oldpasswd;
 		}
 		if (user == NULL) {
-			err = AUTHchangePassword(c, &opwd, &pwd);
+			err = AUTHchangePassword(c, opwd, pwd);
 			if (!enc) {
 				free(pwd);
 				free(opwd);
 			}
 			if (err !=MAL_SUCCEED) {
 				(void) sql_error(m, 02, "ALTER USER: %s", getExceptionMessage(err));
-				GDKfree(err);
+				freeException(err);
 				return (FALSE);
 			}
 		} else {
 			str username = NULL;
-			if ((err = AUTHresolveUser(&username, &c->user)) !=MAL_SUCCEED) {
+			if ((err = AUTHresolveUser(&username, c->user)) !=MAL_SUCCEED) {
 				if (!enc) {
 					free(pwd);
 					free(opwd);
 				}
 				(void) sql_error(m, 02, "ALTER USER: %s", getExceptionMessage(err));
-				GDKfree(err);
+				freeException(err);
 				return (FALSE);
 			}
 			if (strcmp(username, user) == 0) {
 				/* avoid message about changePassword (from MAL level) */
+				GDKfree(username);
 				if (!enc) {
 					free(pwd);
 					free(opwd);
 				}
-				(void) sql_error(m, 02, "ALTER USER: " "use 'ALTER USER SET [ ENCRYPTED ] PASSWORD xxx " "USING OLD PASSWORD yyy' " "when changing your own password");
+				(void) sql_error(m, 02, "ALTER USER: "
+					"use 'ALTER USER SET [ ENCRYPTED ] PASSWORD xxx "
+					"USING OLD PASSWORD yyy' "
+					"when changing your own password");
 				return (FALSE);
 			}
-			err = AUTHsetPassword(c, &user, &pwd);
+			GDKfree(username);
+			err = AUTHsetPassword(c, user, pwd);
 			if (!enc) {
 				free(pwd);
 				free(opwd);
 			}
 			if (err !=MAL_SUCCEED) {
 				(void) sql_error(m, 02, "ALTER USER: %s", getExceptionMessage(err));
-				GDKfree(err);
+				freeException(err);
 				return (FALSE);
 			}
 		}
@@ -323,22 +335,24 @@ monet5_rename_user(ptr _mvc, str olduser, str newuser)
 	sql_table *auths = find_sql_table(sys, "auths");
 	sql_column *auths_name = find_sql_column(auths, "name");
 
-	if ((err = AUTHchangeUsername(c, &olduser, &newuser)) !=MAL_SUCCEED) {
+	if ((err = AUTHchangeUsername(c, olduser, newuser)) !=MAL_SUCCEED) {
 		(void) sql_error(m, 02, "ALTER USER: %s", getExceptionMessage(err));
-		GDKfree(err);
+		freeException(err);
 		return (FALSE);
 	}
 
 	rid = table_funcs.column_find_row(m->session->tr, users_name, olduser, NULL);
 	if (rid == oid_nil) {
-		(void) sql_error(m, 02, "ALTER USER: local inconsistency, " "your database is damaged, user not found in SQL catalog");
+		(void) sql_error(m, 02, "ALTER USER: local inconsistency, "
+				 "your database is damaged, user not found in SQL catalog");
 		return (FALSE);
 	}
 	table_funcs.column_update_value(m->session->tr, users_name, rid, newuser);
 
 	rid = table_funcs.column_find_row(m->session->tr, auths_name, olduser, NULL);
 	if (rid == oid_nil) {
-		(void) sql_error(m, 02, "ALTER USER: local inconsistency, " "your database is damaged, auth not found in SQL catalog");
+		(void) sql_error(m, 02, "ALTER USER: local inconsistency, "
+				 "your database is damaged, auth not found in SQL catalog");
 		return (FALSE);
 	}
 	table_funcs.column_update_value(m->session->tr, auths_name, rid, newuser);
@@ -415,6 +429,7 @@ monet5_user_get_def_schema(mvc *m, int user)
 	if ((rid = table_funcs.column_find_row(m->session->tr, users_name, username, NULL)) != oid_nil)
 		p = table_funcs.column_find_value(m->session->tr, users_schema, rid);
 
+	_DELETE(username);
 	assert(p);
 	schema_id = *(sqlid *) p;
 	_DELETE(p);
@@ -453,8 +468,8 @@ monet5_user_set_def_schema(mvc *m, oid user)
 	if (m->debug &1)
 		fprintf(stderr, "monet5_user_set_def_schema " OIDFMT "\n", user);
 
-	if ((err = AUTHresolveUser(&username, &user)) !=MAL_SUCCEED) {
-		GDKfree(err);
+	if ((err = AUTHresolveUser(&username, user)) !=MAL_SUCCEED) {
+		freeException(err);
 		return (NULL);	/* don't reveal that the user doesn't exist */
 	}
 
