@@ -3,7 +3,7 @@
  * License, v. 2.0.  If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
- * Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.
+ * Copyright 1997 - July 2008 CWI, August 2008 - 2017 MonetDB B.V.
  */
 
 #include <monetdb_config.h>
@@ -196,8 +196,10 @@ exp_op( sql_allocator *sa, list *l, sql_subfunc *f )
 {
 	sql_exp *e = exp_create(sa, e_func);
 	e->card = exps_card(l);
-	if (!l || list_length(l) == 0)
+	if (!l || list_length(l) == 0) 
 		e->card = CARD_ATOM; /* unop returns a single atom */
+	if (f->func->side_effect)
+		e->card = CARD_MULTI;
 	e->l = l;
 	e->f = f; 
 	return e;
@@ -312,25 +314,12 @@ exp_atom_hge(sql_allocator *sa, hge i)
 }
 #endif
 
-sql_exp *
-exp_atom_wrd(sql_allocator *sa, wrd w) 
-{
-	sql_subtype it; 
-
-#ifdef HAVE_HGE
-	sql_find_subtype(&it, "wrd", have_hge ? 18 : 19, 0);
-#else
-	sql_find_subtype(&it, "wrd", 19, 0);
-#endif
-	return exp_atom(sa, atom_int(sa, &it, w ));
-}
-
 sql_exp *
 exp_atom_flt(sql_allocator *sa, flt f) 
 {
 	sql_subtype it; 
 
-	sql_find_subtype(&it, "double", 24, 0);
+	sql_find_subtype(&it, "real", 24, 0);
 	return exp_atom(sa, atom_float(sa, &it, (dbl)f ));
 }
 
@@ -377,15 +366,17 @@ exp_atom_ref(sql_allocator *sa, int i, sql_subtype *tpe)
 }
 
 atom *
-exp_value(sql_exp *e, atom **args, int maxarg)
+exp_value(mvc *sql, sql_exp *e, atom **args, int maxarg)
 {
 	if (!e || e->type != e_atom)
 		return NULL; 
 	if (e->l) {	   /* literal */
 		return e->l;
 	} else if (e->r) { /* param (ie not set) */
+		if (e->flag <= 1) /* global variable */
+			return stack_get_var(sql, e->r); 
 		return NULL; 
-	} else if (e->flag < maxarg) {
+	} else if (sql->emode == m_normal && e->flag < maxarg) { /* do not get the value in the prepared case */
 		return args[e->flag]; 
 	}
 	return NULL; 
@@ -578,8 +569,10 @@ exp_rel(mvc *sql, sql_rel *rel)
 {
 	sql_exp *e = exp_create(sql->sa, e_psm);
 
+	/*
 	rel = rel_optimizer(sql, rel);
 	rel = rel_distribute(sql, rel);
+	*/
 	e->l = rel;
 	e->flag = PSM_REL;
 	return e;
@@ -624,14 +617,26 @@ exp_setrelname(sql_allocator *sa, sql_exp *e, int nr)
 	e->rname = sa_strdup(sa, nme);
 }
 
-sql_exp*
-exp_label(sql_allocator *sa, sql_exp *e, int nr)
+char *
+make_label(sql_allocator *sa, int nr)
 {
 	char name[16], *nme;
 
 	nme = number2name(name, 16, nr);
-	e->name = sa_strdup(sa, nme);
-	e->rname = sa_strdup(sa, nme);
+	return sa_strdup(sa, nme);
+}
+
+sql_exp*
+exp_label(sql_allocator *sa, sql_exp *e, int nr)
+{
+	e->rname = e->name = make_label(sa, nr);
+	return e;
+}
+
+sql_exp*
+exp_label_table(sql_allocator *sa, sql_exp *e, int nr)
+{
+	e->rname = make_label(sa, nr);
 	return e;
 }
 
@@ -788,8 +793,9 @@ exp_match( sql_exp *e1, sql_exp *e2)
 	return 0;
 }
 
+/* c refers to the parent p */
 int 
-exp_refers( sql_exp *c, sql_exp *p)
+exp_refers( sql_exp *p, sql_exp *c)
 {
 	if (c->type == e_column) {
 		if (!p->name || !c->r || strcmp(p->name, c->r) != 0)
@@ -957,8 +963,11 @@ exp_match_exp( sql_exp *e1, sql_exp *e2)
 			if (!subfunc_cmp(e1->f, e2->f) && /* equal functions */
 			    exps_equal(e1->l, e2->l) &&
 			    /* optional order by expressions */
-			    exps_equal(e1->r, e2->r))
-				return 1;
+			    exps_equal(e1->r, e2->r)) {
+				sql_subfunc *f = e1->f;
+				if (!f->func->side_effect)
+					return 1;
+			}
 			break;
 		case e_atom:
 			if (e1->l && e2->l && !atom_cmp(e1->l, e2->l))
@@ -1202,6 +1211,7 @@ rel_find_exp_( sql_rel *rel, sql_exp *e)
 			}
 			return ne;
 		}
+		/* fall through */
 	case e_cmp:	
 	case e_psm:	
 		return NULL;
@@ -1276,6 +1286,32 @@ exp_is_correlation(sql_exp *e, sql_rel *r )
 	return -1;
 }
 
+int
+exp_is_zero(mvc *sql, sql_exp *e) 
+{
+	if (e->type == e_atom) {
+		if (e->l) {
+			return atom_is_zero(e->l);
+		} else if(sql->emode == m_normal && sql->argc > e->flag && EC_COMPUTE(exp_subtype(e)->type->eclass)) {
+			return atom_is_zero(sql->args[e->flag]);
+		}
+	}
+	return 0;
+}
+
+int
+exp_is_not_null(mvc *sql, sql_exp *e) 
+{
+	if (e->type == e_atom) {
+		if (e->l) {
+			return !(atom_null(e->l));
+		} else if(sql->emode == m_normal && sql->argc > e->flag && EC_COMPUTE(exp_subtype(e)->type->eclass)) {
+			return !atom_null(sql->args[e->flag]);
+		}
+	}
+	return 0;
+}
+
 int
 exp_is_atom( sql_exp *e )
 {
@@ -1337,8 +1373,11 @@ exp_has_func( sql_exp *e )
 	case e_convert:
 		return exp_has_func(e->l);
 	case e_func:
-	case e_aggr:
 		return 1;
+	case e_aggr:
+		if (e->l)
+			return exps_has_func(e->l);
+		return 0;
 	case e_cmp:
 		if (e->flag == cmp_or) {
 			return (exps_has_func(e->l) || exps_has_func(e->r));
@@ -1355,6 +1394,43 @@ exp_has_func( sql_exp *e )
 	return 0;
 }
 
+static int
+exps_has_sideeffect( list *exps)
+{
+	node *n;
+	int has_sideeffect = 0;
+
+	for(n=exps->h; n && !has_sideeffect; n=n->next) 
+		has_sideeffect |= exp_has_sideeffect(n->data);
+	return has_sideeffect;
+}
+
+int
+exp_has_sideeffect( sql_exp *e )
+{
+	switch (e->type) {
+	case e_convert:
+		return exp_has_sideeffect(e->l);
+	case e_func:
+		{
+			sql_subfunc *f = e->f;
+
+			if (f->func->side_effect) 
+				return 1;
+			if (e->l)
+				return exps_has_sideeffect(e->l);
+			return 0;
+		}
+	case e_atom:
+	case e_aggr: 
+	case e_cmp:
+	case e_column:
+	case e_psm:
+		return 0;
+	}
+	return 0;
+}
+
 int
 exp_unsafe( sql_exp *e) 
 {
@@ -1403,13 +1479,19 @@ exps_bind_column( list *exps, const char *cname, int *ambiguous )
 			MT_lock_set(&exps->ht_lock);
 			if (!exps->ht && list_length(exps) > HASH_MIN_SIZE) {
 				exps->ht = hash_new(exps->sa, list_length(exps), (fkeyvalue)&exp_key);
-
+				if (exps->ht == NULL) {
+					MT_lock_unset(&exps->ht_lock);
+					return NULL;
+				}
 				for (en = exps->h; en; en = en->next ) {
 					sql_exp *e = en->data;
 					if (e->name) {
 						int key = exp_key(e);
 
-						hash_add(exps->ht, key, e);
+						if (hash_add(exps->ht, key, e) == NULL) {
+							MT_lock_unset(&exps->ht_lock);
+							return NULL;
+						}
 					}
 				}
 			}
@@ -1460,13 +1542,20 @@ exps_bind_column2( list *exps, const char *rname, const char *cname )
 			MT_lock_set(&exps->ht_lock);
 			if (!exps->ht && list_length(exps) > HASH_MIN_SIZE) {
 				exps->ht = hash_new(exps->sa, list_length(exps), (fkeyvalue)&exp_key);
+				if (exps->ht == NULL) {
+					MT_lock_unset(&exps->ht_lock);
+					return NULL;
+				}
 
 				for (en = exps->h; en; en = en->next ) {
 					sql_exp *e = en->data;
 					if (e->name) {
 						int key = exp_key(e);
 
-						hash_add(exps->ht, key, e);
+						if (hash_add(exps->ht, key, e) == NULL) {
+							MT_lock_unset(&exps->ht_lock);
+							return NULL;
+						}
 					}
 				}
 			}
@@ -1520,11 +1609,11 @@ exps_bind_alias( list *exps, const char *rname, const char *cname )
 	return NULL;
 }
 
-int
+unsigned int
 exps_card( list *l ) 
 {
 	node *n;
-	int card = CARD_ATOM;
+	unsigned int card = CARD_ATOM;
 
 	if (l) for(n = l->h; n; n = n->next) {
 		sql_exp *e = n->data;
@@ -1576,21 +1665,21 @@ exps_intern(list *exps)
 }
 
 char *
-compare_func( comp_type t )
+compare_func( comp_type t, int anti )
 {
 	switch(t) {
 	case cmp_equal:
-		return "=";
+		return anti?"<>":"=";
 	case cmp_lt:
-		return "<";
+		return anti?">":"<";
 	case cmp_lte:
-		return "<=";
+		return anti?">=":"<=";
 	case cmp_gte:
-		return ">=";
+		return anti?"<=":">=";
 	case cmp_gt:
-		return ">";
+		return anti?"<":">";
 	case cmp_notequal:
-		return "<>";
+		return anti?"=":"<>";
 	default:
 		return NULL;
 	}
@@ -1744,14 +1833,14 @@ atom *
 exp_flatten(mvc *sql, sql_exp *e) 
 {
 	if (e->type == e_atom) {
-		atom *v =  exp_value(e, sql->args, sql->argc);
+		atom *v =  exp_value(sql, e, sql->args, sql->argc);
 
 		if (v)
 			return atom_dup(sql->sa, v);
 	} else if (e->type == e_convert) {
 		atom *v = exp_flatten(sql, e->l); 
 
-		if (v && atom_cast(v, &e->tpe))
+		if (v && atom_cast(sql->sa, v, &e->tpe))
 			return v;
 		return NULL;
 	} else if (e->type == e_func) {
