@@ -3,7 +3,7 @@
  * License, v. 2.0.  If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
- * Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.
+ * Copyright 1997 - July 2008 CWI, August 2008 - 2017 MonetDB B.V.
  */
 
 #include "monetdb_config.h"
@@ -197,6 +197,7 @@ exp_getdcount( mvc *sql, sql_rel *r , sql_exp *e, lng count)
 	case e_convert:
 		if (e->l)
 			return exp_getdcount(sql, r, e->l, count);
+		/* fall through */
 	case e_func:
 	case e_aggr:
 	case e_atom:
@@ -224,6 +225,7 @@ exp_getranges( mvc *sql, sql_rel *r , sql_exp *e, void **min, void **max)
 	case e_convert:
 		if (e->l)
 			return exp_getranges(sql, r, e->l, min, max);
+		/* fall through */
 	case e_func:
 	case e_aggr:
 	case e_atom:
@@ -237,7 +239,7 @@ static atom *
 exp_getatom( mvc *sql, sql_exp *e, atom *m) 
 {
 	if (is_atom(e->type))
-		return exp_value(e, sql->args, sql->argc);
+		return exp_value(sql, e, sql->args, sql->argc);
 	else if (e->type == e_convert)
 		return exp_getatom(sql, e->l, m);
 	else if (e->type == e_func) {
@@ -434,6 +436,7 @@ rel_getsel(mvc *sql, sql_rel *rel, lng count)
 	case op_project:
 		if (rel->l)
 			return rel_getsel(sql, rel->l, count);
+		/* fall through */
 	default:
 		return 1.0;
 	}
