@@ -4,6 +4,7 @@
 #include "rel_exp.h"
 #include "rel_prop.h"
 #include "rel_remote.h"
+#include "sql_semantic.h"
 #include "sql_mvc.h"
 
 /* we don't name relations directly, but sometimes we need the relation
@@ -53,6 +54,7 @@ rel_destroy_(sql_rel *rel)
 	    is_select(rel->op) ||
 	    is_set(rel->op) ||
 	    rel->op == op_topn ||
+	    rel->op == op_apply ||
 		rel->op == op_sample) {
 		if (rel->l)
 			rel_destroy(rel->l);
@@ -222,6 +224,7 @@ rel_bind_column_(mvc *sql, sql_rel **p, sql_rel *rel, const char *cname )
 		*p = rel;
 		if (rel->l)
 			return rel_bind_column_(sql, p, rel->l, cname);
+		/* fall through */
 	default:
 		return NULL;
 	}
@@ -271,10 +274,18 @@ rel_bind_column2( mvc *sql, sql_rel *rel, const char *tname, const char *cname,
 	} else if (is_set(rel->op) ||
 		   is_sort(rel) ||
 		   is_semi(rel->op) ||
-		   is_apply(rel->op) ||
-		   is_select(rel->op)) {
+		   is_select(rel->op) || 
+		   is_topn(rel->op)) {
 		if (rel->l)
 			return rel_bind_column2(sql, rel->l, tname, cname, f);
+	} else if (is_apply(rel->op)) {
+		sql_exp *e = NULL;
+
+		if (!e && rel->l)
+			e = rel_bind_column2(sql, rel->l, tname, cname, f);
+		if (!e && rel->r && (rel->flag == APPLY_JOIN || rel->flag == APPLY_LOJ))
+			return rel_bind_column2(sql, rel->r, tname, cname, f);
+		return e;
 	}
 	return NULL;
 }
@@ -293,6 +304,7 @@ rel_inplace_setop(sql_rel *rel, sql_rel *l, sql_rel *r, operator_type setop, lis
 	if (l && r)
 		rel->nrcols = l->nrcols + r->nrcols;
 	rel->exps = exps;
+	set_processed(rel);
 	return rel;
 }
 
@@ -358,6 +370,29 @@ rel_setop(sql_allocator *sa, sql_rel *l, sql_rel *r, operator_type setop)
 	return rel;
 }
 
+sql_rel *
+rel_setop_check_types(mvc *sql, sql_rel *l, sql_rel *r, list *ls, list *rs, operator_type op) 
+{
+	list *nls = new_exp_list(sql->sa);
+	list *nrs = new_exp_list(sql->sa);
+	node *n, *m;
+
+	for (n = ls->h, m = rs->h; n && m; n = n->next, m = m->next) {
+		sql_exp *le = n->data;
+		sql_exp *re = m->data;
+
+		if ((rel_convert_types(sql, &le, &re, 1, type_set) < 0))
+			return NULL;
+		append(nls, le);
+		append(nrs, re);
+	}
+	l = rel_project(sql->sa, l, nls);
+	r = rel_project(sql->sa, r, nrs);
+	set_processed(l);
+	set_processed(r);
+	return rel_setop(sql->sa, l, r, op);
+}
+
 sql_rel *
 rel_crossproduct(sql_allocator *sa, sql_rel *l, sql_rel *r, operator_type join)
 {
@@ -415,6 +450,7 @@ rel_label( mvc *sql, sql_rel *r, int all)
 	if (is_project(r->op) && r->exps) {
 		node *ne = r->exps->h;
 
+		r->exps->ht = NULL;
 		for (; ne; ne = ne->next) {
 			if (all) {
 				nr = ++sql->label;
@@ -428,6 +464,7 @@ rel_label( mvc *sql, sql_rel *r, int all)
 		list *exps = r->r;
 		node *ne = exps->h;
 
+		exps->ht = NULL;
 		for (; ne; ne = ne->next) {
 			if (all) {
 				nr = ++sql->label;
@@ -444,8 +481,11 @@ rel_project_add_exp( mvc *sql, sql_rel *rel, sql_exp *e)
 {
 	assert(is_project(rel->op));
 
-	if (!e->rname) 
-		exp_setrelname(sql->sa, e, sql->label);
+	if (!e->rname) {
+		exp_setrelname(sql->sa, e, ++sql->label);
+		if (!e->name)
+			e->name = e->rname;
+	}
 	if (rel->op == op_project) {
 		if (!rel->exps)
 			rel->exps = new_exp_list(sql->sa);
@@ -465,6 +505,8 @@ rel_select_add_exp(sql_allocator *sa, sql_rel *l, sql_exp *e)
 		sql_exp *t = exp_atom_bool(sa, 1);
 		e = exp_compare(sa, e, t, cmp_equal);
 	}
+	if (!l->exps)
+		l->exps = new_exp_list(sa);
 	append(l->exps, e);
 }
 
@@ -630,9 +672,13 @@ rel_basetable(mvc *sql, sql_table *t, const char *atname)
 		for (cn = t->idxs.set->h; cn; cn = cn->next) {
 			sql_exp *e;
 			sql_idx *i = cn->data;
-			sql_subtype *t = sql_bind_localtype("wrd"); /* hash "wrd" */
+			sql_subtype *t = sql_bind_localtype("lng"); /* hash "lng" */
 			char *iname = sa_strconcat( sa, "%", i->base.name);
 
+			/* do not include empty indices in the plan */
+			if (hash_index(i->type) && list_length(i->columns) <= 1)
+				continue;
+
 			if (i->type == join_idx)
 				t = sql_bind_localtype("oid"); 
 
@@ -715,6 +761,21 @@ rel_relational_func(sql_allocator *sa, sql_rel *l, list *exps)
 	return rel;
 }
 
+sql_rel *
+rel_table_func(sql_allocator *sa, sql_rel *l, sql_exp *f, list *exps, int kind)
+{
+	sql_rel *rel = rel_create(sa);
+
+	rel->flag = kind;
+	rel->l = l; /* relation before call */
+	rel->r = f; /* expression (table func call) */
+	rel->op = op_table;
+	rel->exps = exps;
+	rel->card = CARD_MULTI;
+	rel->nrcols = list_length(exps);
+	return rel;
+}
+
 static void
 exps_has_nil(list *exps)
 {
@@ -730,10 +791,10 @@ exps_has_nil(list *exps)
 list *
 rel_projections(mvc *sql, sql_rel *rel, const char *tname, int settname, int intern )
 {
-	int label = sql->label;
-	list *rexps, *exps ;
+	list *lexps, *rexps, *exps;
+	int include_subquery = (intern==2)?1:0;
 
-	if (!rel || (is_subquery(rel) && is_project(rel->op)))
+	if (!rel || (!include_subquery && is_subquery(rel) && rel->op == op_project))
 		return new_exp_list(sql->sa);
 
 	switch(rel->op) {
@@ -762,6 +823,7 @@ rel_projections(mvc *sql, sql_rel *rel, const char *tname, int settname, int int
 	case op_inter:
 		if (rel->exps) {
 			node *en;
+			int label = ++sql->label;
 
 			exps = new_exp_list(sql->sa);
 			for (en = rel->exps->h; en; en = en->next) {
@@ -775,14 +837,18 @@ rel_projections(mvc *sql, sql_rel *rel, const char *tname, int settname, int int
 			}
 			return exps;
 		}
-		exps = rel_projections(sql, rel->l, tname, settname, intern );
-		if (exps) {
-			node *en;
-			for (en = exps->h; en; en = en->next) {
+		lexps = rel_projections(sql, rel->l, tname, settname, intern );
+		rexps = rel_projections(sql, rel->r, tname, settname, intern );
+		exps = sa_list(sql->sa);
+		if (lexps && rexps && exps) {
+			node *en, *ren;
+			int label = ++sql->label;
+			for (en = lexps->h, ren = rexps->h; en && ren; en = en->next, ren = ren->next) {
 				sql_exp *e = en->data;
 				e->card = rel->card;
 				if (!settname) /* noname use alias */
 					exp_setrelname(sql->sa, e, label);
+				append(exps, e);
 			}
 		}
 		return exps;
@@ -836,6 +902,7 @@ rel_bind_path_(sql_rel *rel, sql_exp *e, list *path )
 			assert(0);
 			break;
 		}
+		/* fall through */
 	case op_groupby:
 	case op_project:
 	case op_table:
@@ -1035,39 +1102,63 @@ rel_push_join(mvc *sql, sql_rel *rel, sql_exp *ls, sql_exp *rs, sql_exp *rs2, sq
 }
 
 sql_rel *
-rel_or(mvc *sql, sql_rel *l, sql_rel *r, list *oexps, list *lexps, list *rexps)
+rel_or(mvc *sql, sql_rel *rel, sql_rel *l, sql_rel *r, list *oexps, list *lexps, list *rexps)
 {
-	sql_rel *rel, *ll = l->l, *rl = r->l;
+	sql_rel *ll = l->l, *rl = r->l;
+	list *ls, *rs;
 
-	if (l == r && is_outerjoin(l->op)) { /* merge both lists */
+	assert(!lexps || l == r);
+	if (l == r && lexps) { /* merge both lists */
 		sql_exp *e = exp_or(sql->sa, lexps, rexps);
 		list *nl = oexps?oexps:new_exp_list(sql->sa); 
 		
 		rel_destroy(r);
 		append(nl, e);
+		if (is_outerjoin(l->op) && is_processed(l)) 
+			l = rel_select(sql->sa, l, NULL);
 		l->exps = nl;
 		return l;
 	}
 
-	if (l->op == r->op && 
-		((ll == rl && l->r == r->r && l->r == NULL /* or check if columns are equal*/) ||
-
-		(exps_card(l->exps) == exps_card(r->exps) && exps_card(l->exps) <= CARD_ATOM))) {
+	/* favor or expressions over union */
+	if (l->op == r->op && l->op == op_select &&
+	    ll == rl && ll == rel && !rel_is_ref(l) && !rel_is_ref(r)) {
 		sql_exp *e = exp_or(sql->sa, l->exps, r->exps);
 		list *nl = new_exp_list(sql->sa); 
 		
 		rel_destroy(r);
 		append(nl, e);
 		l->exps = nl;
+
+		/* merge and expressions */
+		ll = l->l;
+		while (ll && ll->op == op_select && !rel_is_ref(ll)) {
+			list_merge(l->exps, ll->exps, (fdup)NULL);
+			l->l = ll->l;
+			ll->l = NULL;
+			rel_destroy(ll);
+			ll = l->l;
+		}
 		return l;
 	}
-	l = rel_project(sql->sa, l, rel_projections(sql, l, NULL, 1, 1));
-	r = rel_project(sql->sa, r, rel_projections(sql, r, NULL, 1, 1));
+
+	if (rel) {
+		ls = rel_projections(sql, rel, NULL, 1, 1);
+		rs = rel_projections(sql, rel, NULL, 1, 1);
+	} else {
+		ls = rel_projections(sql, l, NULL, 1, 1);
+		rs = rel_projections(sql, r, NULL, 1, 1);
+	}
 	set_processed(l);
 	set_processed(r);
-	rel = rel_setop(sql->sa, l, r, op_union);
+	rel = rel_setop_check_types(sql, l, r, ls, rs, op_union);
+	if (!rel)
+		return NULL;
 	rel->exps = rel_projections(sql, rel, NULL, 1, 1);
+	set_processed(rel);
 	rel = rel_distinct(rel);
+	if (!rel)
+		return NULL;
 	if (exps_card(l->exps) <= CARD_AGGR &&
 	    exps_card(r->exps) <= CARD_AGGR)
 	{
@@ -1130,3 +1221,35 @@ rel_add_identity(mvc *sql, sql_rel *rel, sql_exp **exp)
 		return rel;
 	return _rel_add_identity(sql, rel, exp);
 }
+
+sql_exp *
+rel_find_column( sql_allocator *sa, sql_rel *rel, const char *tname, const char *cname )
+{
+	if (!rel)
+		return NULL;
+
+	if (rel->exps && (is_project(rel->op) || is_base(rel->op))) {
+		int ambiguous = 0;
+		sql_exp *e = exps_bind_column2(rel->exps, tname, cname);
+		if (!e && cname[0] == '%')
+			e = exps_bind_column(rel->exps, cname, &ambiguous);
+		if (e && !ambiguous)
+			return exp_alias(sa, e->rname, exp_name(e), e->rname, cname, exp_subtype(e), e->card, has_nil(e), is_intern(e));
+	}
+	if (is_project(rel->op) && rel->l) {
+		return rel_find_column(sa, rel->l, tname, cname);
+	} else if (is_join(rel->op)) {
+		sql_exp *e = rel_find_column(sa, rel->l, tname, cname);
+		if (!e)
+			e = rel_find_column(sa, rel->r, tname, cname);
+		return e;
+	} else if (is_set(rel->op) ||
+		   is_sort(rel) ||
+		   is_semi(rel->op) ||
+		   is_select(rel->op)) {
+		if (rel->l)
+			return rel_find_column(sa, rel->l, tname, cname);
+	}
+	return NULL;
+}
+
