@@ -3,7 +3,7 @@
  * License, v. 2.0.  If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
- * Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.
+ * Copyright 1997 - July 2008 CWI, August 2008 - 2017 MonetDB B.V.
  */
 
 #ifndef BAT_LOGGER_H
@@ -14,7 +14,7 @@
 
 extern logger *bat_logger;
 
-extern int bat_logger_init( logger_functions *lf );
-extern int bat_logger_init_shared( logger_functions *lf );
+extern void bat_logger_init( logger_functions *lf );
+extern void bat_logger_init_shared( logger_functions *lf );
 
 #endif /*BAT_LOGGER_H */
