@@ -3,7 +3,7 @@
  * License, v. 2.0.  If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
- * Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.
+ * Copyright 1997 - July 2008 CWI, August 2008 - 2017 MonetDB B.V.
  */
 
 #include "monetdb_config.h"
@@ -16,21 +16,17 @@ _delta_cands(sql_trans *tr, sql_table *t)
 {
 	sql_column *c = t->columns.set->h->data;
 	/* create void,void bat with length and oid's set */
-	BAT *tids = BATnew(TYPE_void, TYPE_void, 0, TRANSIENT);
+	BAT *tids = COLnew(0, TYPE_void, 0, TRANSIENT);
 	size_t nr = store_funcs.count_col(tr, c, 1);
 
 	if (!tids)
 		return NULL;
-	tids->H->seq = 0;
-	tids->T->seq = 0;
+	tids->tseqbase = 0;
 	BATsetcount(tids, (BUN) nr);
-	tids->H->revsorted = 0;
-	tids->T->revsorted = 0;
+	tids->trevsorted = 0;
 
-	tids->T->key = 1;
-	tids->T->dense = 1;
-	tids->H->key = 1;
-	tids->H->dense = 1;
+	tids->tkey = 1;
+	tids->tdense = 1;
 
 	if (store_funcs.count_del(tr, t)) {
 		BAT *d, *diff = NULL;
@@ -85,11 +81,18 @@ delta_full_bat_( sql_column *c, sql_delta *bat, int temp)
 		b = i;
 	} else {
 		if (BATcount(i)) {
-			assert(b->htype == TYPE_void);
 			r = COLcopy(b, b->ttype, 1, TRANSIENT); 
 			bat_destroy(b); 
+			if (r == NULL) {
+				bat_destroy(i);
+				return NULL;
+			}
 			b = r;
-			BATappend(b, i, TRUE); 
+			if (BATappend(b, i, NULL, TRUE) != GDK_SUCCEED) {
+				bat_destroy(b);
+				bat_destroy(i);
+				return NULL;
+			}
 			needcopy = 0;
 		}
 		bat_destroy(i); 
@@ -99,12 +102,16 @@ delta_full_bat_( sql_column *c, sql_delta *bat, int temp)
 		uv = temp_descriptor(bat->uvbid);
 		if (ui && BATcount(ui)) {
 			if (needcopy) {
-				assert(b->htype == TYPE_void);
 				r = COLcopy(b, b->ttype, 1, TRANSIENT); 
 				bat_destroy(b); 
 				b = r;
 			}
-			void_replace_bat(b, ui, uv, TRUE);
+			if (void_replace_bat(b, ui, uv, TRUE) == BUN_NONE) {
+				bat_destroy(ui);
+				bat_destroy(uv);
+				bat_destroy(b);
+				return NULL;
+			}
 		}
 		bat_destroy(ui); 
 		bat_destroy(uv); 
@@ -142,38 +149,6 @@ full_destroy(sql_column *c, BAT *b)
 		bat_destroy(b);
 }
 
-static oid
-column_lookup_row(sql_trans *tr, sql_column *c, const void *value) 
-{
-	BAT *b = NULL, *s = NULL;
-	oid rid = oid_nil;
-
-	b = full_column(tr, c);
-	if (!b)
-		return oid_nil;
-
-	if (store_funcs.count_del(tr, c->t)) 
-		s = store_funcs.bind_del(tr, c->t, RD_INS);
-
-	if (BAThash(b, 0) == GDK_SUCCEED) {
-		BATiter cni = bat_iterator(b);
-		BUN p;
-
-		HASHloop(cni, cni.b->T->hash, p, value) {
-			oid pos = p;
-
-			if (!s || BUNfnd(s, &pos) == BUN_NONE) {
-				rid = p;
-				break;
-			}
-		}
-	}
-	if (s)
-		bat_destroy(s);
-	full_destroy(c, b);
-	return rid;
-}
-
 static oid
 column_find_row(sql_trans *tr, sql_column *c, const void *value, ...)
 {
@@ -183,9 +158,6 @@ column_find_row(sql_trans *tr, sql_column *c, const void *value, ...)
 	sql_column *n = NULL;
 
 	va_start(va, value);
-	if ((n = va_arg(va, sql_column *)) == NULL) 
-		return column_lookup_row(tr, c, value);
-
 	s = delta_cands(tr, c->t);
 	if (!s)
 		return oid_nil;
@@ -198,7 +170,7 @@ column_find_row(sql_trans *tr, sql_column *c, const void *value, ...)
 	bat_destroy(s);
 	s = r;
 	full_destroy(c, b);
-	do {
+	while ((n = va_arg(va, sql_column *)) != NULL) {
 		value = va_arg(va, void *);
 		c = n;
 
@@ -211,7 +183,7 @@ column_find_row(sql_trans *tr, sql_column *c, const void *value, ...)
 		bat_destroy(s);
 		s = r;
 		full_destroy(c, b);
-	} while ((n = va_arg(va, sql_column *)) != NULL); 
+	}
 	va_end(va);
 	if (BATcount(s) == 1) {
 		BATiter ri = bat_iterator(s);
@@ -229,8 +201,12 @@ column_find_value(sql_trans *tr, sql_column *c, oid rid)
 	void *res = NULL;
 
 	b = full_column(tr, c);
-	if (b)
-		q = BUNfnd(BATmirror(b), (ptr) &rid);
+	if (b) {
+		if (rid < b->hseqbase || rid >= b->hseqbase + BATcount(b))
+			q = BUN_NONE;
+		else
+			q = rid - b->hseqbase;
+	}
 	if (q != BUN_NONE) {
 		BATiter bi = bat_iterator(b);
 		void *r;
@@ -238,6 +214,7 @@ column_find_value(sql_trans *tr, sql_column *c, oid rid)
 
 		res = BUNtail(bi, q);
 		sz = ATOMlen(b->ttype, res);
+		// FIXME unchecked_malloc GDKmalloc can return NULL
 		r = GDKmalloc(sz);
 		memcpy(r,res,sz);
 		res = r;
@@ -252,7 +229,7 @@ column_update_value(sql_trans *tr, sql_column *c, oid rid, void *value)
 	assert(rid != oid_nil);
 
 	store_funcs.update_col(tr, c, &rid, value, c->type.type->localtype);
-	return 0;
+	return LOG_OK;
 }
 
 static int
@@ -264,8 +241,7 @@ table_insert(sql_trans *tr, sql_table *t, ...)
 	int cnt = 0;
 
 	va_start(va, t);
-	for (val = va_arg(va, void *); n && val; n = n->next, val = va_arg(va, void *))
-	{
+	for (; n && (val = va_arg(va, void *)) != NULL; n = n->next) {
 		sql_column *c = n->data;
 		store_funcs.append_col(tr, c, val, c->type.type->localtype);
 		cnt++;
@@ -274,9 +250,9 @@ table_insert(sql_trans *tr, sql_table *t, ...)
 	if (n) {
 		fprintf(stderr, "called table_insert(%s) with wrong number of args (%d,%d)\n", t->base.name, list_length(t->columns.set), cnt);
 		assert(0);
-		return -1;
+		return LOG_ERR;
 	}
-	return 0;
+	return LOG_OK;
 }
 
 static int
@@ -285,20 +261,21 @@ table_delete(sql_trans *tr, sql_table *t, oid rid)
 	assert(rid != oid_nil);
 
 	store_funcs.delete_tab(tr, t, &rid, TYPE_oid);
-	return 0;
+	return LOG_OK;
 }
 
 
 
 /* returns table rids, for the given select ranges */
 static rids *
-rids_select( sql_trans *tr, sql_column *key, void *key_value_low, void *key_value_high, ...)
+rids_select( sql_trans *tr, sql_column *key, const void *key_value_low, const void *key_value_high, ...)
 {
 	va_list va;
 	BAT *b = NULL, *r = NULL, *s = NULL;
 	rids *rs = ZNEW(rids);
-	void *kvl = key_value_low, *kvh = key_value_high;
-	int hi = 0;
+	const void *kvl = key_value_low, *kvh = key_value_high;
+	/* if pointers are equal, make it an inclusive select */
+	int hi = key_value_low == key_value_high;
 
 	s = delta_cands(tr, key->t);
 	b = full_column(tr, key);
@@ -346,11 +323,16 @@ rids_orderby(sql_trans *tr, rids *r, sql_column *orderby_col)
 	b = full_column(tr, orderby_col);
 	s = BATproject(r->data, b);
 	full_destroy(orderby_col, b);
-	BATsort(NULL, &o, NULL, s, NULL, NULL, 0, 0);
+	if (BATsort(NULL, &o, NULL, s, NULL, NULL, 0, 0) != GDK_SUCCEED) {
+		bat_destroy(s);
+		return NULL;
+	}
 	bat_destroy(s);
 	s = BATproject(o, r->data);
-	bat_destroy(r->data);
 	bat_destroy(o);
+	if (s == NULL)
+		return NULL;
+	bat_destroy(r->data);
 	r->data = s;
 	return r;
 }
@@ -386,13 +368,18 @@ static rids *
 rids_join(sql_trans *tr, rids *l, sql_column *lc, rids *r, sql_column *rc)
 {
 	BAT *lcb, *rcb, *s = NULL, *d = NULL;
+	gdk_return ret;
 	
 	lcb = full_column(tr, lc);
 	rcb = full_column(tr, rc);
-	BATjoin(&s, &d, lcb, rcb, l->data, r->data, FALSE, BATcount(lcb));
+	ret = BATjoin(&s, &d, lcb, rcb, l->data, r->data, FALSE, BATcount(lcb));
 	bat_destroy(l->data);
 	bat_destroy(d);
-	l->data = s;
+	if (ret != GDK_SUCCEED) {
+		l->data = NULL;
+	} else {
+		l->data = s;
+	}
 	full_destroy(lc, lcb);
 	full_destroy(rc, rcb);
 	return l;
@@ -402,44 +389,90 @@ static subrids *
 subrids_create(sql_trans *tr, rids *t1, sql_column *rc, sql_column *lc, sql_column *obc)
 {
 	/* join t1.rc with lc order by obc */
-	subrids *r = ZNEW(subrids);
+	subrids *r;
 	BAT *lcb, *rcb, *s, *obb, *d = NULL, *o, *g, *ids, *rids = NULL;
-	
+	gdk_return ret;
+
 	lcb = full_column(tr, lc);
 	rcb = full_column(tr, rc);
-
 	s = delta_cands(tr, lc->t);
-	BATjoin(&rids, &d, lcb, rcb, s, t1->data, FALSE, BATcount(lcb));
-	bat_destroy(d);
+	if (lcb == NULL || rcb == NULL || s == NULL) {
+		if (lcb)
+			full_destroy(rc, lcb);
+		if (rcb)
+			full_destroy(rc, rcb);
+		bat_destroy(s);
+		return NULL;
+	}
+
+	ret = BATjoin(&rids, &d, lcb, rcb, s, t1->data, FALSE, BATcount(lcb));
 	bat_destroy(s);
 	full_destroy(rc, rcb);
+	if (ret != GDK_SUCCEED) {
+		full_destroy(rc, lcb);
+		return NULL;
+	}
+	bat_destroy(d);
 
 	s = BATproject(rids, lcb);
 	full_destroy(lc, lcb);
+	if (s == NULL) {
+		bat_destroy(rids);
+		return NULL;
+	}
 	lcb = s;
 
-	obb = full_column(tr, obc);
+	if ((obb = full_column(tr, obc)) == NULL) {
+		bat_destroy(lcb);
+		bat_destroy(rids);
+		return NULL;
+	}
 	s = BATproject(rids, obb);
 	full_destroy(obc, obb);
+	if (s == NULL) {
+		bat_destroy(lcb);
+		bat_destroy(rids);
+		return NULL;
+	}
 	obb = s;
 
 	/* need id, obc */
 	ids = o = g = NULL;
-	BATsort(&ids, &o, &g, lcb, NULL, NULL, 0, 0);
+	ret = BATsort(&ids, &o, &g, lcb, NULL, NULL, 0, 0);
 	bat_destroy(lcb);
+	if (ret != GDK_SUCCEED) {
+		bat_destroy(obb);
+		bat_destroy(rids);
+		return NULL;
+	}
 
 	s = NULL;
-	BATsort(NULL, &s, NULL, obb, o, g, 0, 0);
+	ret = BATsort(NULL, &s, NULL, obb, o, g, 0, 0);
 	bat_destroy(obb);
 	bat_destroy(o);
 	bat_destroy(g);
+	if (ret != GDK_SUCCEED) {
+		bat_destroy(ids);
+		bat_destroy(rids);
+		return NULL;
+	}
 
 	o = BATproject(s, rids);
 	bat_destroy(rids);
 	bat_destroy(s);
+	if (o == NULL) {
+		bat_destroy(ids);
+		return NULL;
+	}
 	rids = o;
 
 	assert(ids->ttype == TYPE_int && ATOMtype(rids->ttype) == TYPE_oid);
+	r = ZNEW(subrids);
+	if (r == NULL) {
+		bat_destroy(ids);
+		bat_destroy(rids);
+		return NULL;
+	}
 	r->id = 0;
 	r->pos = 0;
 	r->ids = ids;
@@ -487,6 +520,7 @@ rids_diff(sql_trans *tr, rids *l, sql_column *lc, subrids *r, sql_column *rc )
 {
 	BAT *lcb = full_column(tr, lc), *s, *d, *rids, *diff;
 	BAT *rcb = full_column(tr, rc);
+	gdk_return ret;
 
 	s = BATproject(r->rids, rcb);
 	full_destroy(rc, rcb);
@@ -496,18 +530,54 @@ rids_diff(sql_trans *tr, rids *l, sql_column *lc, subrids *r, sql_column *rc )
 
 	diff = BATdiff(s, rcb, NULL, NULL, 0, BUN_NONE);
 
-	BATjoin(&rids, &d, lcb, s, NULL, diff, FALSE, BATcount(s));
+	ret = BATjoin(&rids, &d, lcb, s, NULL, diff, FALSE, BATcount(s));
 	bat_destroy(diff);
 	bat_destroy(d);
 	full_destroy(lc, lcb);
 	bat_destroy(s);
 
 	bat_destroy(l->data);
-	l->data = rids;
+	if (ret != GDK_SUCCEED) {
+		l->data = NULL;
+	}
+	else {
+		l->data = rids;
+	}
 	return l;
 }
 
-int 
+static int
+table_vacuum(sql_trans *tr, sql_table *t)
+{
+	BAT *tids = delta_cands(tr, t);
+	BAT **cols;
+	node *n;
+
+	if (!tids)
+		return SQL_ERR;
+	// FIXME unchecked_malloc NEW_ARRAY can return NULL
+	cols = NEW_ARRAY(BAT*, cs_size(&t->columns));
+	for (n = t->columns.set->h; n; n = n->next) {
+		sql_column *c = n->data;
+		BAT *v = store_funcs.bind_col(tr, c, RDONLY);
+
+		if (!v)
+			return SQL_ERR;
+		cols[c->colnr] = BATproject(tids, v);
+		BBPunfix(v->batCacheid);
+	}
+	sql_trans_clear_table(tr, t);
+	for (n = t->columns.set->h; n; n = n->next) {
+		sql_column *c = n->data;
+
+		store_funcs.append_col(tr, c, cols[c->colnr], TYPE_bat);
+		BBPunfix(cols[c->colnr]->batCacheid);
+	}
+	_DELETE(cols);
+	return LOG_OK;
+}
+
+void
 bat_table_init( table_functions *tf )
 {
 	tf->column_find_row = column_find_row;
@@ -516,6 +586,7 @@ bat_table_init( table_functions *tf )
 	tf->column_update_value = column_update_value;
 	tf->table_insert = table_insert;
 	tf->table_delete = table_delete;
+	tf->table_vacuum = table_vacuum;
 	
 	tf->rids_select = rids_select;
 	tf->rids_orderby = rids_orderby;
@@ -529,5 +600,4 @@ bat_table_init( table_functions *tf )
 	tf->subrids_nextid = subrids_nextid;
 	tf->subrids_destroy = subrids_destroy;
 	tf->rids_diff = rids_diff;
-	return LOG_OK;
 }
